# -*- coding: utf-8 -*-
"""
conftest.py: example pytest configuration for tdda.referencetest

Source repository: http://github.com/tdda/tdda

License: MIT

Copyright (c) Stochastic Solutions Limited 2016-2019
"""

from tdda.referencetest.pytestconfig import (pytest_addoption,
                                             pytest_collection_modifyitems,
                                             set_default_data_location,
                                             ref)




# accounts_verify_25k_against_1k.py

from __future__ import print_function
import pandas as pd

from tdda.constraints.pd.constraints import verify_df

df = pd.read_csv('testdata/accounts25k.csv')
v = verify_df(df, 'accounts1k.tdda')
vdf = v.to_frame()
print(vdf)




# -*- coding: utf-8 -*-

"""This module provides the
:py:class:`~tdda.referencetest.referencetestcase.ReferenceTestCase` class,
which extends the
standard :py:class:`unittest.TestCase` test-case class, augmenting it
with methods for checking correctness of files against reference data.

It also provides a main() function, which can be used to run (and
regenerate) reference tests which have been implemented using subclasses
of ReferenceTestCase.

For example::

    from tdda.referencetest import ReferenceTestCase
    import my_module

    class TestMyClass(ReferenceTestCase):
        def test_my_csv_function(self):
            result = my_module.my_csv_function(self.tmp_dir)
            self.assertCSVFileCorrect(result, 'result.csv')

        def test_my_pandas_dataframe_function(self):
            result = my_module.my_dataframe_function()
            self.assertDataFrameCorrect(result, 'result.csv')

        def test_my_table_function(self):
            result = my_module.my_table_function()
            self.assertStringCorrect(result, 'table.txt', kind='table')

        def test_my_graph_function(self):
            result = my_module.my_graph_function()
            self.assertStringCorrect(result, 'graph.txt', kind='graph')

    TestMyClass.set_default_data_location('testdata')

    if __name__ == '__main__':
        ReferenceTestCase.main()

Tagged Tests
~~~~~~~~~~~~

If the tests are run with the ``--tagged`` or ``-1`` (the digit one)
command-line option, then only tests that have been decorated with
``referencetest.tag``, are run. This is a mechanism for allowing
only a chosen subset of tests to be run, which is useful during
development. The ``@tag`` decorator can be applied to either test
classes or test methods.

If the tests are run with the ``--istagged`` or ``-0`` (the digit
zero) command-line option, then no tests are run; instead, the
framework reports the full module names of any test classes that have
been decorated with ``@tag``, or which contain any tests that have been
decorated with ``@tag``.

For example::

    from tdda.referencetest import ReferenceTestCase, tag
    import my_module

    class TestMyClass1(ReferenceTestCase):
        @tag
        def test_a(self):
            ...

        def test_b(self):
            ...

    @tag
    class TestMyClass2(ReferenceTestCase):
        def test_x(self):
            ...

        def test_y(self):
            ...

If run with ``python mytests.py --tagged``, only the tagged tests are
run (``TestMyClass1.test_a``, ``TestMyClass2.test_x`` and
``TestMyClass2.test_y``).

Regeneration of Results
~~~~~~~~~~~~~~~~~~~~~~~

When its main is run with ``--write-all`` or ``--write`` (or ``-W`` or ``-w``
respectively), it causes the framework to regenerate reference data
files. Different kinds of reference results can be regenerated by
passing in a comma-separated list of *kind* names immediately after
the ``--write`` option. If no list of *kind* names is provided, then all
test results will be regenerated.

To regenerate all reference results (or generate them for the first time)

.. code-block:: bash

   pytest -s --write-all

To regenerate just a particular kind of reference (e.g. table results)

.. code-block:: bash

    python my_tests.py --write table

To regenerate a number of different kinds of reference (e.g. both table
and graph results)

.. code-block:: bash

    python my_tests.py --write table graph

``unittest`` Integration Details
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import sys
import unittest

from tdda.referencetest.referencetest import ReferenceTest, tag


class ReferenceTestCase(unittest.TestCase, ReferenceTest):
    """
    Wrapper around the
    :py:class:`~tdda.referencetest.referencetest.ReferenceTest`
    class to allow it to operate as a test-case class using the
    ``unittest`` testing framework.

    The ``ReferenceTestCase`` class is a mix-in of
    :py:class:`unittest.TestCase`
    and :py:class:`~tdda.referencetest.referencetest.ReferenceTest`,
    so it can be used as the base class for unit tests, allowing the
    tests to use any of the standard ``unittest`` *assert* methods,
    and also use any of the ``referencetest`` *assert* extensions.
    """
    tag = tag

    def __init__(self, *args, **kwargs):
        # Initializer for a ReferenceTestCase instance.
        # This is called automatically by the ``unittest`` runtime framework.
        unittest.TestCase.__init__(self, *args, **kwargs)
        ReferenceTest.__init__(self, self.assertTrue)

    @staticmethod
    def main(module=None, argv=None):
        """
        Wrapper around the :py:func:`unittest.main()` entry point.

        This is the same as the :py:func:`~tdda.referencetestcase.main()`
        function, and is provided just as a convenience, as it means that
        tests using the ``ReferenceTestCase`` class only need to import
        that single class on its own.
        """
        argv, tagged, check = _set_flags_from_argv(argv)
        _run_tests(module=module, argv=argv, tagged=tagged, check=check)


def _run_tests(module=None, argv=None, tagged=False, check=False):
    """
    Run tests
    """
    if argv is None:
        argv = sys.argv
    loader = (TaggedTestLoader(check) if tagged or check
              else unittest.defaultTestLoader)
    if module is None:
        unittest.main(argv=argv, testLoader=loader)
    else:
        unittest.main(module=module, argv=argv, testLoader=loader)


def _set_flags_from_argv(argv=None):
    """
    This is used to set the class's write flag if a **-write** or
    **--write-all** (or **-W** or **-w**) option is passed on the
    command line, either using the argv provided,
    or :py:data:`sys.argv` otherwise.

    The **-write-all** (or **-W**) option takes no parameters, and turns on
    reference-regeneration for all kinds of results.

    The **-write** (or **-w**) option takes a list of parameter, consisting
    of names of result kinds to be regenerated. The names can be separate
    parameters, or they can be a single comma-separated parameter.

    The framework reports on each file being regenerated, by default. Use
    The **-wquiet** option to make it rewrite files quietly.

    The **--wquiet** option causes files to be rewritten silently.

    If the **-1** or **--tagged** option is set, then only run tagged tests.

    If the **-0** or **--check** option is set, then only report tagged tests,
    without running them.

    A tuple is returned, containing ``argv`` or :py:data:`sys.argv`, with
    any of the *rewrite*, *tagged* or *check* options options removed,
    plus the *tagged* and *checked* state.
    """
    if argv is None:
        argv = sys.argv

    tagged = False
    check = False
    regenerate = False

    for i, arg in enumerate(argv[1:]):
        if arg.startswith('-') and not arg.startswith('--'):
            for flag in arg[1:]:
                if flag == 'W':
                    regenerate = True
                    arg = arg.replace('W', '')
                elif flag == '1':
                    tagged = True
                    arg = arg.replace('1', '')
                elif flag == '0':
                    check = True
                    arg = arg.replace('0', '')
            if arg == '-':
                argv = argv[:i+1] + argv[i+2:]
            else:
                argv = argv[:i+1] + [arg] + argv[i+2:]
        else:
            break

    for quietflag in ('-wquiet', '--wquiet'):
        if quietflag in argv:
            idx = argv.index(quietflag)
            ReferenceTestCase.set_defaults(verbose=False)
            argv = argv[:idx] + argv[idx+1:]

    for writeflag in ('--W', '--write-all'):
        if writeflag in argv:
            idx = argv.index(writeflag)
            if idx:
                regenerate = True
                argv = argv[:idx] + argv[idx+1:]
                break

    for writeflag in ('-w', '--w', '--write'):
        if writeflag in argv:
            idx = argv.index(writeflag)
            if idx:
                if idx < len(argv) - 1:
                    for r in argv[idx+1:]:
                        for kind in r.split(','):
                            ReferenceTestCase.set_regeneration(kind)
                else:
                    raise Exception('--write option requires parameters; '
                                    'use --write-all to regenerate all '
                                    'reference results')
            argv = argv[:idx]
            break

    for option in ('--tagged', '--istagged'):
        if option in argv:
            idx = argv.index(option)
            if idx:
                argv = argv[:idx] + argv[idx+1:]
                if option in ('-0', '--istagged'):
                    check = True
                else:
                    tagged = True

    if regenerate:
        ReferenceTestCase.set_regeneration()

    return (argv, tagged, check)


class TaggedTestLoader(unittest.TestLoader):
    """
    Subclass of TestLoader, which strips out any non-tagged tests.
    """
    def __init__(self, check, printer=None):
        unittest.TestLoader.__init__(self)
        self.check = check
        self.print = printer or print

    def loadTestsFromTestCase(self, *args, **kwargs):
        suite = unittest.TestLoader.loadTestsFromTestCase(self, *args,
                                                          **kwargs)
        return self._tagged_tests_only(suite)

    def loadTestsFromModule(self, *args, **kwargs):
        suite = unittest.TestLoader.loadTestsFromModule(self, *args, **kwargs)
        return self._tagged_tests_only(suite)

    def loadTestsFromName(self, *args, **kwargs):
        suite = unittest.TestLoader.loadTestsFromName(self, *args, **kwargs)
        return self._tagged_tests_only(suite)

    def loadTestsFromNames(self, *args, **kwargs):
        suite = unittest.TestLoader.loadTestsFromNames(self, *args, **kwargs)
        return self._tagged_tests_only(suite)

    def _tagged_tests_only(self, suite):
        newsuite = unittest.TestSuite()
        cases = set()
        for test in suite:
            if isinstance(test, unittest.suite.TestSuite):
                test = self._tagged_tests_only(test)
            if self.check and not isinstance(test, unittest.suite.TestSuite):
                cases.add('%s.%s' % (test.__class__.__module__,
                                     test.__class__.__name__))
            else:
                newsuite.addTest(test)
        if self.check:
            for module in cases:
                self.print(module)
        return newsuite

    def getTestCaseNames(self, testCaseClass):
        names = unittest.TestLoader.getTestCaseNames(self, testCaseClass)
        if hasattr(testCaseClass, '_tagged'):
            return names
        else:
            return [name for name in names
                         if hasattr(getattr(testCaseClass, name), '_tagged')]


def main():
    """
    Wrapper around the :py:func:`unittest.main()` entry point.
    """
    ReferenceTestCase.main()




from __future__ import print_function
from artists.giacometti.utils import Timer
t = Timer()
N = (10 * 1000 * 1000)
a = [None] * N
print(t.SplitTimeStr('[None] * N: '))
a = [1] * N
print(t.SplitTimeStr('[1] * N: '))
a = [None for i in range(N)]
print(t.SplitTimeStr('[None for i in range(N)]: '))
a = ['' for i in range(N)]
print(t.SplitTimeStr("['' for i in range(N)]: "))
a = [1 for i in range(N)]
print(t.SplitTimeStr("[1 for i in range(N)]: "))




# -*- coding: utf-8 -*-

"""
Standard configuration boilerplate for use with pytest.

The conftest.py configuration file in a test suite should enable reference
tests by including the following two lines:

    from tdda.referencetest.pytestconfig import *

    set_default_data_location('/directory/where/my/reference/test/data/lives')
"""

import pytest

from tdda.referencetest import referencepytest

def pytest_addoption(parser):
    """
    Extend pytest to include the --write, --write-all regeneration
    command-line options, and the --tagged and --istagged tagging options.
    """
    referencepytest.addoption(parser)


def pytest_collection_modifyitems(session, config, items):
    """
    Extend pytest to only run tagged tests if run with --tagged,
    and to report tagged tests if run with --istagged (and -s).
    """
    referencepytest.tagged(config, items)


@pytest.fixture(scope='module')
def ref(request):
    """
    Declare a fixture called 'ref' which will inject a ReferenceTest
    instance into a test. It's automatically configured to regenerate
    results if the ``--write-all`` or ``--write``  options are used on
    the command line.
    """
    return referencepytest.ref(request)


def set_default_data_location(location, kind=None):
    """
    Expose the referencepytest.set_default_data_location function here,
    so that it's easy to use from a conftest.py thst imports this file.
    """
    referencepytest.set_default_data_location(location, kind=kind)




# -*- coding: utf-8 -*-

"""
checkfiles.py: comparison mechanism for text files

Source repository: http://github.com/tdda/tdda

License: MIT

Copyright (c) Stochastic Solutions Limited 2016-2018
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
# from __future__ import unicode_literals

import os
import re
import sys
import tempfile


class BaseComparison(object):
    """
    Common base class for different implementations of comparisons.
    """
    def __init__(self, print_fn=None, verbose=True, tmp_dir=None):
        """
        Constructor for an instance of the BaseComparison class.

        The optional print_fn parameter is a function to be used to
        display information while comparison operations are running.
        If specified, it should be a function with the same signature
        as python's builtin (__future__) print function.
        """
        self.print_fn = print_fn
        self.verbose = verbose
        self.tmp_dir = tmp_dir or tempfile.gettempdir()

    def info(self, msgs, s):
        """
        Add an item to the list of messages, and also display it immediately
        if verbose is set.
        """
        if s is not None:
            msgs.append(s)
            if self.verbose and self.print_fn:
                self.print_fn(s)

    @staticmethod
    def compare_with(actual, expected, qualifier=None, binary=False):
        qualifier = '' if not qualifier else (qualifier + ' ')
        if os.path.exists(expected):
            if binary:
                return None
            else:
                msg = 'Compare %swith:\n    %s %s %s\n'
                cmd = diffcmd()
        else:
            msg = 'Initialize %sfrom actual content with:\n    %s %s %s'
            cmd = copycmd()
        return msg % (qualifier, cmd, actual, expected)


class Diffs(object):
    """
    Class for representing a list of messages describing differences
    resulting from applying comparisons to a set of pairs of actual/expected
    files or strings.

    The 'messages' are a stream of message-strings, whereas the
    'reconstructions' are a list of one-per-comparison objects.

    It doesn't (currently) try to tie up the messages to individual
    comparison operations.
    """
    def __init__(self, lines=None):
        self.lines = lines or []
        self.reconstructions = []

    def append(self, line):
        self.lines.append(line)

    def add_reconstruction(self, r):
        self.reconstructions.append(r)

    def message(self):
        return '\n'.join(self.lines)

    def __repr__(self):
        # representation of a Diffs object is just its lines; used in tests.
        return repr(self.lines)

    def __eq__(self, other):
        # comparison between a Diffs object and a list of messages just
        # compares the lines part of the diffs; used in tests.
        if isinstance(other, Diffs):
            return self.lines == other.lines
        elif type(other) is list:
            return self.lines == other
        else:
            return False

    def __iter__(self):
        # iterating over a Diffs object is the same as iterating over
        # its internal messages; used in tests.
        return iter(self.lines)


def diffcmd():
    return 'fc' if os.name and os.name != 'posix' else 'diff'


def copycmd():
    return 'copy' if os.name and os.name != 'posix' else 'cp'




# -*- coding: utf-8 -*-
"""
test_using_referencepytest.py: A simple example of how to use
                               tdda.referencetest.referencepytest with pytest

Source repository: http://github.com/tdda/tdda

License: MIT

Copyright (c) Stochastic Solutions Limited 2019
"""
import time

from generators import generate_string


def testZero():
    assert True


def testOne():
    time.sleep(1)
    assert 1 == 1


def testExampleStringGeneration(ref):
    actual = generate_string()
    ref.assertStringCorrect(actual, 'expected.html')


def testTwo():
    time.sleep(2)
    assert 2 == 2


def testThree():
    time.sleep(3)
    assert 3 == 3



# -*- coding: utf-8 -*-

#
# Unit tests for string functions from tdda.referencetest.checkfiles
#

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import division

import unittest

from tdda.referencetest.checkfiles import FilesComparison


class TestInternals(unittest.TestCase):
    def test_diff_marker(self):
        compare = FilesComparison()
        self.assertEqual(compare.diff_marker('ABC', 'XYZ'), '(ABC|XYZ)')
        self.assertEqual(compare.diff_marker('ABC:', 'ABC: yes'), 'ABC:(| yes)')
        self.assertEqual(compare.diff_marker('', 'AAA'), '(|AAA)')
        self.assertEqual(compare.diff_marker('AAA', ''), '(AAA|)')
        self.assertEqual(compare.diff_marker('ABC', 'AXC'), 'A(B|X)C')

    def test_single_pattern(self):
        compare = FilesComparison()
        cpatterns = compare.compile_patterns(['gr.*t'])
        self.assertTrue(compare.check_patterns(cpatterns,
                                               'great', 'grapefruit'))

    def test_unanchored_patterns(self):
        compare = FilesComparison()
        cpatterns = compare.compile_patterns(['A\d{2}B', 'X[a-z]+Y'])
        for actual, expected in [('A22BC', 'A99BC'),
                                 ('XappleY', 'XtrafficY'),
                                 ('A22BXappleY', 'A99BXtrafficY'),
                                 ('froggyA22BXappleY', 'froggyA99BXtrafficY'),
                                 ('frA22BXappleYoggy', 'frA99BXtrafficYoggy'),
                                 ('A22BA99B', 'A99BA22B')]:
            self.assertTrue(compare.check_patterns(cpatterns,
                                                   actual, expected),
                            '%s <--> %s' % (actual, expected))
        for actual, expected in [('A222BC', 'A99BC'),
                                 ('222BC', 'A99BC'),
                                 ('XappleYXappleY', 'XappleY')]:
            self.assertFalse(compare.check_patterns(cpatterns,
                                                    actual, expected),
                             '%s <--> %s' % (actual, expected))

    def test_anchored_patterns(self):
        compare = FilesComparison()
        cpatterns = compare.compile_patterns([r'^\d+$'])
        for actual, expected in [('2', '222222222222'),
                                 ('2', '2'),
                                 ('02', '2'),
                                 ('2', '23'),
                                 ('123', '564')]:
            self.assertTrue(compare.check_patterns(cpatterns,
                                                   actual, expected),
                            '%s <--> %s' % (actual, expected))
        for actual, expected in [('2', '222222222222a22'),
                                 ('', '23'),
                                 ('123', '')]:
            self.assertFalse(compare.check_patterns(cpatterns,
                                                    actual, expected),
                             '%s <--> %s' % (actual, expected))

    def test_grouped_pattern(self):
        compare = FilesComparison()
        cpatterns = compare.compile_patterns(['(a|an) (grapefruit|apple)'])
        self.assertTrue(compare.check_patterns(cpatterns,
                                               'a grapefruit', 'an apple'))
        self.assertTrue(compare.check_patterns(cpatterns,
                                               'I have a grapefruit', 'I have an apple'))
        self.assertTrue(compare.check_patterns(cpatterns,
                                               'I have a grapefruit and an apple',
                                               'I have an apple and a grapefruit'))
        self.assertFalse(compare.check_patterns(cpatterns,
                                               'I have a grapefruit and a banana',
                                               'I have an apple and a grapefruit'))


class TestStrings(unittest.TestCase):
    def test_strings_ok(self):
        compare = FilesComparison()
        self.assertEqual(compare.check_strings([], []), (0, []))
        self.assertEqual(compare.check_strings(['abc'], ['abc']), (0, []))
        self.assertEqual(compare.check_strings(['ab', 'c'], ['ab', 'c']),
                         (0, []))

    def test_strings_fail(self):
        compare = FilesComparison()
        self.assertEqual(compare.check_strings([], ['x'], create_temporaries=False),
                         (1, ['Strings have different numbers of lines, '
                              'differences start at end of actual string',
                              'No files available for comparison']))
        self.assertEqual(compare.check_strings(['y'], ['x'], create_temporaries=False),
                         (1, ['1 line is different, starting at line 1',
                              'No files available for comparison']))

    def test_print(self):
        msgs = []
        compare = FilesComparison(print_fn=lambda x: msgs.append(x))
        compare.check_strings(['a'], ['b'], create_temporaries=False)
        self.assertEqual(msgs, ['1 line is different, starting at line 1',
                                'No files available for comparison'])

    def test_strip(self):
        compare = FilesComparison()
        self.assertEqual(compare.check_strings(['   abc'], ['abc'],
                                               create_temporaries=False),
                         (1, ['1 line is different, starting at line 1',
                              'No files available for comparison']))
        self.assertEqual(compare.check_strings(['   abc'], ['abc'],
                                               lstrip=True), (0, []))
        self.assertEqual(compare.check_strings(['abc   '], ['abc'],
                                               rstrip=True), (0, []))
        self.assertEqual(compare.check_strings(['   abc   '], ['abc'],
                                               lstrip=True, rstrip=True),
                         (0, []))

    def test_ignore_substrings(self):
        compare = FilesComparison()
        self.assertEqual(compare.check_strings(['abc','red', 'banana'],
                                               ['abc','blue', 'grapefruit'],
                                               create_temporaries=False),
                         (1, ['2 lines are different, starting at line 2',
                              'No files available for comparison']))
        self.assertEqual(compare.check_strings(['abc','blue', 'banana'],
                                               ['abc','red', 'grapefruit'],
                                               ignore_substrings=['re'],
                                               create_temporaries=False),
                         (1, ['1 line is different, starting at line 3',
                              'No files available for comparison',
                              'Note exclusions:',
                              '    ignore_substrings:',
                              '        re']))
        self.assertEqual(compare.check_strings(['abc','red', 'banana'],
                                               ['abc','blue', 'grapefruit'],
                                               ignore_substrings=['ue','gra']),
                         (0, []))

    def test_ignore_patterns(self):
        compare = FilesComparison()

        # red != blue, banana != grapefruit => 2 failures
        self.assertEqual(compare.check_strings(['abc','red', 'banana'],
                                               ['abc','blue', 'grapefruit'],
                                               create_temporaries=False),
                         (1, ['2 lines are different, starting at line 2',
                              'No files available for comparison']))

        # red != blue, banana !~ gr.*t => 2 failures
        self.assertEqual(compare.check_strings(['abc','red', 'banana'],
                                               ['abc','blue', 'grapefruit'],
                                               ignore_patterns=['gr.*t'],
                                               create_temporaries=False),
                         (1, ['2 lines are different, starting at line 2',
                              'No files available for comparison',
                              'Note exclusions:',
                              '    ignore_patterns:',
                              '        gr.*t']))

        # red != blue, but great DOES ~ gr.*t => 1 failure
        self.assertEqual(compare.check_strings(['abc','red', 'great'],
                                               ['abc','blue', 'grapefruit'],
                                               ignore_patterns=['gr.*t'],
                                               create_temporaries=False),
                         (1, ['1 line is different, starting at line 2',
                              'No files available for comparison',
                              'Note exclusions:',
                              '    ignore_patterns:',
                              '        gr.*t']))

        # spangle DOES ~ sp......, and breadfruit DOES ~ .*fruit => success
        self.assertEqual(compare.check_strings(['abc','spangle', 'breadfruit'],
                                               ['abc','spanner', 'grapefruit'],
                                               ignore_patterns=[
                                                   'sp.....',
                                                   '[bg].*fruit'
                                               ]),
                         (0, []))

    def test_preprocess(self):
        compare = FilesComparison()
        def strip_first_five(strings):
            return [s[5:] for s in strings]
        def strip_first_seven(strings):
            return [s[7:] for s in strings]
        self.assertEqual(compare.check_strings(['abc','spangle', 'breadfruit'],
                                               ['abc','spanner', 'grapefruit'],
                                               preprocess=strip_first_five,
                                               create_temporaries=False),
                         (1, ['1 line is different, starting at line 2',
                              'No files available for comparison']))
        self.assertEqual(compare.check_strings(['abc','spangle', 'breadfruit'],
                                               ['abc','spanner', 'grapefruit'],
                                               preprocess=strip_first_seven),
                         (0, []))

    def test_permutations(self):
        compare = FilesComparison()
        self.assertEqual(compare.check_strings(['abc','spangle', 'spanner'],
                                               ['spangle','spanner', 'abc'],
                                               max_permutation_cases=1,
                                               create_temporaries=False),
                         (1, ['3 lines are different, starting at line 1',
                              'No files available for comparison']))
        self.assertEqual(compare.check_strings(['abc','spangle', 'spanner'],
                                               ['abc','spanner', 'spangle'],
                                               max_permutation_cases=2),
                         (0, []))
        self.assertEqual(compare.check_strings(['abc','spangle', 'spanner'],
                                               ['spangle','spanner', 'abc'],
                                               max_permutation_cases=3),
                         (0, []))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""
Helpers for command-line option flags for discover and verify
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import argparse
import sys


DISCOVER_HELP = '''
Optional flags are:

  * -r or --rex
      Include regular expression generation. Disabled by default.
  * -R or --norex
      Exclude regular expression generation (the default)
'''

VERIFY_HELP = '''
Optional flags are:

  * -a, --all
      Report all fields, even if there are no failures
  * -f, --fields
      Report only fields with failures
  * -7, --ascii
      Report in ASCII form, without using special characters.
  * --epsilon E
      Use this value of epsilon for fuzziness in comparing numeric values.
'''

DETECT_HELP = '''
Optional flags are:

  * -a, --all
      Report all fields, even if there are no failures
  * -f, --fields
      Report only fields with failures
  * -7, --ascii
      Report in ASCII form, without using special characters.
  * --epsilon E
      Use this value of epsilon for fuzziness in comparing numeric values.
  * --write-all
      Include passing records in the output.
  * --per-constraint
      Write one additional column per failing constraint, to show if a
      constraint passed or failed, as well as the n_failures
      total column for each row. This is set by default.
  * --no-per-constraint
      Disable the --per-constraint flag, so that the only constraint-based
      column written out is the nfailures field.
  * --no-original-fields
      Do not write out any of the original columns. By default, all of the
      original columns are written out, unless you use --output-fields.
  * --output-fields FIELD1 FIELD2 ...
      Specify original columns to write out.
  * --interleave
      In the output, place the verification fields immediately after
      the original field to which they correspond.
  * --index
      Include a row-number index in the output file.
      The row number is automatically included if no output fields are
      specified. Rows are usually numbered from 1, unless the (feather)
      input file already has an index.

'''

def discover_parser(usage=''):
    formatter = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(prog='tdda discover',
                                     epilog=usage + DISCOVER_HELP,
                                     formatter_class=formatter)
    parser.add_argument('-?', '--?', action='help',
                        help='same as -h or --help')
    parser.add_argument('-r', '--rex', action='store_true',
                        help='include regular expression generation')
    parser.add_argument('-R', '--norex', action='store_true',
                        help='exclude regular expression generation')
    parser.add_argument('-7', '--ascii', action='store_true',
                        help='report without using special characters')
    return parser


def discover_flags(parser, args, params):
    flags, more = parser.parse_known_args(args)
    if len(more) > 0:
        print(parser.epilog, file=sys.stderr)
        sys.exit(1)
    params['inc_rex'] = flags.rex
    return flags


def verify_parser(usage=''):
    formatter = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(prog='tdda verify',
                                     epilog=usage + VERIFY_HELP,
                                     formatter_class=formatter)
    parser.add_argument('-?', '--?', action='help',
                        help='same as -h or --help')
    parser.add_argument('-a', '--all', action='store_true',
                        help='report all fields, even if there are '
                             'no failures')
    parser.add_argument('-f', '--fields', action='store_true',
                        help='report only fields with failures')
    parser.add_argument('-7', '--ascii', action='store_true',
                        help='report without using special characters')
    parser.add_argument('-t', '--type_checking', choices=['strict', 'sloppy'],
                        help='"sloppy" means consider all numeric types '
                             'equivalent')
    parser.add_argument('-epsilon', '--epsilon', type=float,
                        help='epsilon fuzziness')
    return parser


def detect_parser(usage=''):
    formatter = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(prog='tdda detect',
                                     epilog=usage + DETECT_HELP,
                                     formatter_class=formatter)
    parser.add_argument('-?', '--?', action='help',
                        help='same as -h or --help')
    parser.add_argument('-a', '--all', action='store_true',
                        help='report all fields, even if there are '
                             'no failures')
    parser.add_argument('-f', '--fields', action='store_true',
                        help='report only fields with failures')
    parser.add_argument('-7', '--ascii', action='store_true',
                        help='report without using special characters')
    parser.add_argument('-t', '--type_checking', choices=['strict', 'sloppy'],
                        help='"sloppy" means consider all numeric types '
                             'equivalent')
    parser.add_argument('-epsilon', '--epsilon', type=float,
                        help='epsilon fuzziness')
    parser.add_argument('--write-all', action='store_true',
                        help='Include passing records')
    parser.add_argument('--per-constraint', action='store_true',
                        help='Write one flag column per failing constraint '
                             'in addition to n_failures. Set by default.')
    parser.add_argument('--no-per-constraint', action='store_true',
                        help='Do not write out any per-constraint flag columns')
    parser.add_argument('--no-output-fields', action='store_true',
                        help='Do not write out any original fields in the '
                             'output. By default, all original columns will '
                             'be included.')
    parser.add_argument('--output-fields', nargs='*',
                        help='Specify original columns to write out.')
    parser.add_argument('--interleave', action='store_true',
                        help='Interleave ok columns with original fields.')
    parser.add_argument('--index', action='store_true',
                        help='Include a row-number index in the output file '
                             'when detecting. Rows are usually numbered from '
                             '1, unless the input file already has an index.')
    parser.add_argument('--int', dest='boolean_ints', action='store_true',
                        help='Write out boolean fields as integers, with '
                             '1 for true and 0 for false.')
    return parser


def verify_flags(parser, args, params):
    flags, more = parser.parse_known_args(args)
    if len(more) > 0:
        print('Unexpected arguments %s\n' % ' '.join(more),
              parser.epilog, file=sys.stderr)
        sys.exit(1)
    params.update({
        'report': 'all',
        'ascii': False,
    })
    if flags.all:
        params['report'] = 'all'
    elif flags.fields:
        params['report'] = 'fields'
    if flags.ascii:
        params['ascii'] = True
    if flags.type_checking is not None:
        params['type_checking'] = flags.type_checking
    if flags.epsilon is not None:
        params['epsilon'] = float(flags.epsilon)
    return flags


def detect_flags(parser, args, params):
    flags, more = parser.parse_known_args(args)
    if len(more) > 0:
        print(parser.epilog, file=sys.stderr)
        sys.exit(1)
    params.update({
        'report': 'records',
        'ascii': False,
    })
    if flags.per_constraint and flags.no_per_constraint:
        print('You must not specify both --per-constraint and '
              '--no-per-constraint.', file=sys.stderr)
        sys.exit(1)
    if flags.output_fields and flags.no_output_fields:
        print('You must not specify both --output-fields and '
              '--no-output-fields.', file=sys.stderr)
        sys.exit(1)
    if flags.ascii:
        params['ascii'] = True
    if flags.type_checking is not None:
        params['type_checking'] = flags.type_checking
    if flags.epsilon is not None:
        params['epsilon'] = float(flags.epsilon)
    if flags.write_all:
        params['write_all'] = True
    if not flags.no_per_constraint:
        params['per_constraint'] = True
    if flags.index:
        params['index'] = True
    if flags.boolean_ints:
        params['boolean_ints'] = True

    if flags.output_fields is not None:
        params['output_fields'] = flags.output_fields
    elif not flags.no_output_fields:
        params['output_fields'] = []

    if flags.interleave:
        params['interleave'] = True
    params['in_place'] = False  # Only applicable in API case
    params['report'] = 'records'
    return flags




# -*- coding: utf-8 -*-

"""
TDDA constraint discovery and verification, common underlying functionality.
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import datetime
import re
import sys

from collections import OrderedDict

from tdda.constraints.base import (
    PRECISIONS,
    SIGNS,
    STANDARD_FIELD_CONSTRAINTS,
    verify, detect,
    native_definite,
    DatasetConstraints,
    FieldConstraints,
    Verification,
    TypeConstraint,
    MinConstraint, MaxConstraint, SignConstraint,
    MinLengthConstraint, MaxLengthConstraint,
    NoDuplicatesConstraint, MaxNullsConstraint,
    AllowedValuesConstraint, RexConstraint,
    EPSILON_DEFAULT,
    fuzzy_greater_than, fuzzy_less_than
)

from tdda.constraints.extension import (BaseConstraintCalculator,
                                        BaseConstraintDetector)

if sys.version_info[0] >= 3:
    unicode_string = str
    byte_string = bytes
    long_type = int
else:
    unicode_string = unicode
    byte_string = str
    long_type = long

DEBUG = False

TYPE_CHECKING_OPTIONS = ('strict', 'sloppy')
DEFAULT_TYPE_CHECKING = 'sloppy'

MAX_CATEGORIES = 20     # String fields with up to 20 categories will
                        # generate AllowedValues constraints


class BaseConstraintVerifier(BaseConstraintCalculator, BaseConstraintDetector):
    """
    The :py:mod:`BaseConstraintVerifier` class provides a generic
    framework for verifying constraints.

    A concrete implementation of this class is constructed by creating
    a mix-in subclass which inherits both from :py:mod:`BaseConstraintVerifier`
    and from specific implementations of :py:mod:`BaseConstraintCalculator`
    and :py:mod:`BaseConstraintDetector`.
    """
    def __init__(self, epsilon=None, type_checking=None, **kwargs):
        self.epsilon = EPSILON_DEFAULT if epsilon is None else epsilon
        self.type_checking = type_checking or DEFAULT_TYPE_CHECKING
        assert self.type_checking in TYPE_CHECKING_OPTIONS
        self.cache = {}

    def verifiers(self):
        """
        Returns a dictionary mapping constraint types to their callable
        (bound) verification methods.
        """
        return {
            'type': self.verify_tdda_type_constraint,
            'min': self.verify_min_constraint,
            'max': self.verify_max_constraint,
            'min_length': self.verify_min_length_constraint,
            'max_length': self.verify_max_length_constraint,
            'sign': self.verify_sign_constraint,
            'max_nulls': self.verify_max_nulls_constraint,
            'no_duplicates': self.verify_no_duplicates_constraint,
            'allowed_values': self.verify_allowed_values_constraint,
            'rex': self.verify_rex_constraint,
        }

    def verify(self, constraints, VerificationClass=Verification, **kwargs):
        """
        Apply verifiers to a set of constraints, for reporting
        """
        return verify(constraints, self.get_column_names(), self.verifiers(),
                      VerificationClass=VerificationClass,
                      detected_records_writer=self.write_detected_records,
                      **kwargs)

    def detect(self, constraints, VerificationClass=Verification,
               outpath=None, write_all=False, per_constraint=False,
               output_fields=None, index=False, in_place=False,
               rownumber_is_index=True, boolean_ints=False, **kwargs):
        """
        Apply verifiers to a set of constraints, for detection.

        Note that if there is a constraint for a field that does not exist,
        then it fails verification, but there are no records to detect
        against. Similarly if the field exists but the dataset has no
        records.
        """
        return detect(constraints, self.get_column_names(), self.verifiers(),
                      VerificationClass=VerificationClass,
                      detect_outpath=outpath, detect_write_all=write_all,
                      detect_per_constraint=per_constraint,
                      detect_output_fields=output_fields,
                      detect_index=index,
                      detect_in_place=in_place,
                      detected_records_writer=self.write_detected_records,
                      rownumber_is_index=rownumber_is_index,
                      boolean_ints=boolean_ints,
                      **kwargs)

    def get_cached_value(self, value, colname, f):
        """
        Return cached value of colname, calculating it and caching it
        first, if it is not already there.
        """
        col_cache = self.cache_values(colname)
        if not value in col_cache:
            col_cache[value] = f(colname)
        return col_cache[value]

    def cache_values(self, colname):
        """
        Returns the dictionary for colname from the cache, first creating
        it if there isn't one on entry.
        """
        if not colname in self.cache:
            self.cache[colname] = {}
        return self.cache[colname]

    def verify_min_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given column satisfies the minimum value
        constraint specified.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        precision = getattr(constraint, 'precision', 'fuzzy') or 'fuzzy'
        assert precision in PRECISIONS

        if self.is_null(value):   # a null minimum is not considered to be an
            return True           # active constraint, so is always satisfied

        m = self.get_min(colname)
        if self.is_null(m):       # If there are no values, no value can
            return True           # the minimum constraint

        if (isinstance(value, datetime.datetime)
                or isinstance(value, datetime.date)):
            m = self.to_datetime(m)

        if not self.types_compatible(m, value):
            result = False
        elif (precision == 'closed' or isinstance(value, datetime.datetime)
                                    or isinstance(value, datetime.date)):
            result = m >= value
        elif precision == 'open':
            result = m > value
        else:
            result = fuzzy_greater_than(m, value, self.epsilon)

        if detect and not bool(result):
            self.detect_min_constraint(colname, value, precision, self.epsilon)
        return result

    def verify_max_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given column satisfies the maximum value
        constraint specified.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        precision = getattr(constraint, 'precision', 'fuzzy') or 'fuzzy'
        assert precision in PRECISIONS

        if self.is_null(value):   # a null maximum is not considered to be an
            return True           # active constraint, so is always satisfied

        M = self.get_max(colname)
        if self.is_null(M):       # If there are no values, no value can
            return True           # the maximum constraint

        if (isinstance(value, datetime.datetime)
                or isinstance(value, datetime.date)):
            M = self.to_datetime(M)

        if not self.types_compatible(M, value):
            result = False
        elif (precision == 'closed' or isinstance(value, datetime.datetime)
                                    or isinstance(value, datetime.date)):
            result = M <= value
        elif precision == 'open':
            result = M < value
        else:
            result = fuzzy_less_than(M, value, self.epsilon)

        if detect and not bool(result):
            self.detect_max_constraint(colname, value, precision, self.epsilon)
        return result

    def verify_min_length_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given (string) column satisfies the minimum length
        constraint specified.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        if self.is_null(value):   # a null minimum length is not considered
            return True           # to be an active constraint, so is always
                                  # satisfied

        if self.get_tdda_type(colname) != 'string':
            return False

        m = self.get_min_length(colname)
        if self.is_null(m):       # If there are no values, no value can
            return True           # the minimum length constraint

        result = m >= value

        if detect and not bool(result):
            self.detect_min_length_constraint(colname, value)
        return result

    def verify_max_length_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given (string) column satisfies the maximum length
        constraint specified.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        if self.is_null(value):   # a null minimum length is not considered
            return True           # to be an active constraint, so is always
                                  # satisfied

        if self.get_tdda_type(colname) != 'string':
            return False

        M = self.get_max_length(colname)
        if self.is_null(M):       # If there are no values, no value can
            return True           # the maximum length constraint

        result = M <= value

        if detect and not bool(result):
            self.detect_max_length_constraint(colname, value)
        return result

    def verify_tdda_type_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given column satisfies the supplied type constraint.
        """
        if not self.column_exists(colname):
            return False

        required_type = constraint.value
        allowed_types = (required_type if type(required_type) in (list, tuple)
                         else [required_type])
        if len(allowed_types) == 1 and self.is_null(allowed_types[0]):
            return True  # a null type is not considered to be an
                         # active constraint, so is always satisfied

        actual_type = self.get_tdda_type(colname)
        if self.type_checking == 'strict':
            result = actual_type in allowed_types
        else:
            if actual_type in allowed_types:
                result = True       # definitely OK if the types actually match
            elif 'int' in allowed_types and actual_type == 'real':
                result = self.get_non_integer_values_count(colname) == 0
            elif 'bool' in allowed_types and actual_type == 'real':
                result = self.get_non_integer_values_count(colname) == 0
            elif 'bool' in allowed_types and actual_type == 'string':
                # boolean columns with nulls get converted to dtype
                # object, which is usually used for strings
                result =  self.get_all_non_nulls_boolean(colname)
            else:
                result = False

        if detect and not bool(result):
            self.detect_tdda_type_constraint(colname, required_type)
        return result

    def verify_sign_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given column satisfies the supplied sign constraint.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        if self.is_null(value):   # a null value (as opposed to the string
                                  # 'null') is not considered to be an
            return True           # active constraint, so is always satisfied

        m = self.get_min(colname)
        M = self.get_max(colname)
        if self.is_null(m):
            return True  # no values: cannot violate constraint

        if type(m) not in (bool, int, long_type, float):
            result = False
        elif type(M) not in (bool, int, long_type, float):
            result = False
        elif value == 'null':
             result = False
        elif value == 'positive':
            result = m > 0
        elif value == 'non-negative':
            result = m >= 0
        elif value == 'zero':
            result = m == M == 0
        elif value == 'non-positive':
            result = M <= 0
        elif value == 'negative':
            result = M < 0
        assert value in SIGNS

        if detect and not bool(result):
            self.detect_sign_constraint(colname, value)
        return result

    def verify_max_nulls_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given column satisfies the supplied constraint
        that it should contain no nulls.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        if self.is_null(value):   # a null value is not considered to be an
            return True           # active constraint, so is always satisfied
        result = self.get_null_count(colname) <= value

        if bool(result) or not detect:
            return result
        self.detect_max_nulls_constraint(colname, value)
        return False

    def verify_no_duplicates_constraint(self, colname, constraint,
                                        detect=False):
        """
        Verify whether a given column satisfies the constraint supplied,
        that it should contain no duplicate (non-null) values.
        """
        if not self.column_exists(colname):
            return False

        value = constraint.value
        if value is False or self.is_null(value):
            # a null value is not considered to be an active constraint,
            # so is always satisfied
            return True

        assert value == True      # value not really used; but should be True

        non_nulls = self.get_non_null_count(colname)
        result = self.get_nunique(colname) == non_nulls

        if detect and not bool(result):
            self.detect_no_duplicates_constraint(colname, value)
        return result

    def verify_allowed_values_constraint(self, colname, constraint,
                                         detect=False):
        """
        Verify whether a given column satisfies the constraint on allowed
        (string) values provided.
        """
        if not self.column_exists(colname):
            return False

        exclusions = self.allowed_values_exclusions()
        allowed_values = constraint.value
        if allowed_values is None:      # a null value is not considered
            return True                 # to be an active constraint,
                                        # so is always satisfied

        n_allowed_values = len(allowed_values)
        n_actual_values = self.get_nunique(colname)

        if not detect and n_actual_values > n_allowed_values:
            # can know the result without actually identifying values
            result = False
        else:
            actual_values = self.get_unique_values(colname)
            exclusions = exclusions or []

            violations = (set(actual_values) - set(allowed_values)
                                             - set(exclusions))
            result = len(violations) == 0

        if detect and not bool(result):
            self.detect_allowed_values_constraint(colname, allowed_values,
                                                  violations)
        return result

    def verify_rex_constraint(self, colname, constraint, detect=False):
        """
        Verify whether a given column satisfies a given regular
        expression constraint (by matching at least one of the regular
        expressions given).
        """
        if not self.column_exists(colname):
            return False
        if self.get_tdda_type(colname) != 'string':
            return False

        violations = self.calc_rex_constraint(colname, constraint,
                                              detect=detect)
        if bool(violations):
            # a truthy result means some values failed the constraint
            if detect:
                self.detect_rex_constraint(colname, violations)
            return False
        else:
            return True

    def get_min(self, colname):
        """Looks up cached minimum of column, or calculates and caches it"""
        return self.get_cached_value('min', colname, self.calc_min)

    def get_max(self, colname):
        """Looks up cached maximum of column, or calculates and caches it"""
        return self.get_cached_value('max', colname, self.calc_max)

    def get_min_length(self, colname):
        """
        Looks up cached minimum string length in column,
        or calculates and caches it
        """
        return self.get_cached_value('min_length', colname,
                                     self.calc_min_length)

    def get_max_length(self, colname):
        """
        Looks up cached maximum string length in column,
        or calculates and caches it
        """
        return self.get_cached_value('max_length', colname,
                                     self.calc_max_length)

    def get_tdda_type(self, colname):
        """
        Looks up cached tdda type of a column,
        or calculates and caches it
        """
        return self.get_cached_value('tdda_type', colname, self.calc_tdda_type)

    def get_null_count(self, colname):
        """
        Looks up or caches the number of nulls in a column,
        or calculates and caches it
        """
        return self.get_cached_value('null_count', colname,
                                     self.calc_null_count)

    def get_non_null_count(self, colname):
        """
        Looks up or caches the number of non-null values in a column,
        or calculates and caches it
        """
        return self.get_cached_value('non_null_count', colname,
                                     self.calc_non_null_count)

    def get_nunique(self, colname):
        """
        Looks up or caches the number of unique (distinct) values in a column,
        or calculates and caches it.
        """
        return self.get_cached_value('nunique', colname, self.calc_nunique)

    def get_unique_values(self, colname):
        """
        Looks up or caches the list of unique (distinct) values in a column,
        or calculates and caches it.
        """
        return self.get_cached_value('uniques', colname,
                                     self.calc_unique_values)

    def get_non_integer_values_count(self, colname):
        """
        Looks up or caches the number of non-integer values in a real column,
        or calculates and caches it.
        """
        return self.get_cached_value('non_integer_values_count', colname,
                                     self.calc_non_integer_values_count)

    def get_all_non_nulls_boolean(self, colname):
        """
        Looks up or caches the number of non-integer values in a real column,
        or calculates and caches it.
        """
        return self.get_cached_value('all_non_nulls_boolean', colname,
                                     self.calc_all_non_nulls_boolean)


class BaseConstraintDiscoverer(BaseConstraintCalculator):
    """
    The :py:mod:`BaseConstraintDiscoverer` class provides a generic
    framework for discovering constraints.

    A concrete implementation of this class is constructed by creating
    a mix-in subclass which inherits both from :py:mod:`BaseConstraintDiscover`
    and from a specific implementation of :py:mod:`BaseConstraintCalculator`.
    """
    def __init__(self, inc_rex=False, seed=None, **kwargs):
        self.inc_rex = inc_rex
        self.seed = seed

    def discover(self):
        field_constraints = []
        for col in self.get_column_names():
            constraints = self.discover_field_constraints(col)
            if constraints:
                field_constraints.append(constraints)
        if field_constraints:
            return DatasetConstraints(field_constraints)
        else:
            return None

    def discover_field_constraints(self, fieldname):
        min_constraint = max_constraint = None
        min_length_constraint = max_length_constraint = None
        sign_constraint = no_duplicates_constraint = None
        max_nulls_constraint = allowed_values_constraint = None
        rex_constraint = None

        type_ = self.calc_tdda_type(fieldname)
        if type_ == 'other':
            return None         # Unrecognized or complex
        else:
            type_constraint = TypeConstraint(type_)
        length = self.get_nrecords()

        if length > 0:  # Things are not very interesting when there is no data
            nNull = self.calc_null_count(fieldname)
            nNonNull = self.calc_non_null_count(fieldname)
            assert nNull + nNonNull == length
            if nNull < 2:
                max_nulls_constraint = MaxNullsConstraint(nNull)

            # Useful info:
            uniqs = None
            n_unique = -1   # won't equal number of non-nulls later on
            if type_ in ('string', 'int'):
                n_unique = self.calc_nunique(fieldname)
                if type_ == 'string':
                    if n_unique <= MAX_CATEGORIES:
                        uniqs = self.calc_unique_values(fieldname,
                                                        include_nulls=False)
                    if uniqs:
                        avc = AllowedValuesConstraint(uniqs)
                        allowed_values_constraint = avc

            if nNonNull > 0:
                if type_ == 'string':
                    # We don't generate a min, max or sign constraints for
                    # strings. But we do generate min and max length
                    # constraints
                    if (uniqs is None and n_unique > 0):
                        # There were too many for us to have bothered getting
                        # them all before, but we need them now.
                        uniqs = self.calc_unique_values(fieldname,
                                                        include_nulls=False)
                    if uniqs:
                        if type(uniqs[0]) is unicode_string:
                            L = [len(v) for v in uniqs]
                        else:
                            L = [len(v.decode('UTF-8')) for v in uniqs]
                        m = min(L)
                        M = max(L)
                        min_length_constraint = MinLengthConstraint(m)
                        max_length_constraint = MaxLengthConstraint(M)
                else:
                    # Non-string fields all potentially get min and max values
                    m = self.calc_min(fieldname)
                    M = self.calc_max(fieldname)
                    if not self.is_null(m):
                        min_constraint = MinConstraint(m)
                    if not self.is_null(M):
                        max_constraint = MaxConstraint(M)

                    # Non-date fields potentially get a sign constraint too.
                    if min_constraint and max_constraint and type_ != 'date':
                        if m == M == 0:
                            sign_constraint = SignConstraint('zero')
                        elif m >= 0:
                            sign = 'positive' if m > 0 else 'non-negative'
                            sign_constraint = SignConstraint(sign)
                        elif M <= 0:
                            sign = 'negative' if M < 0 else 'non-positive'
                            sign_constraint = SignConstraint(sign)
                        # else:
                            # mixed
                    elif self.is_null(m) and type_ != 'date':
                        sign_constraint = SignConstraint('null')

            if n_unique == nNonNull and n_unique > 1 and type_ != 'real':
                # all values are unique
                no_duplicates_constraint = NoDuplicatesConstraint()

        if type_ == 'string' and self.inc_rex:
            rex_constraint = RexConstraint(self.find_rexes(fieldname,
                                                           values=uniqs,
                                                           seed=self.seed))

        constraints = [c for c in [type_constraint,
                                   min_constraint, max_constraint,
                                   min_length_constraint,
                                   max_length_constraint,
                                   sign_constraint, max_nulls_constraint,
                                   no_duplicates_constraint,
                                   allowed_values_constraint,
                                   rex_constraint]
                         if c is not None]
        return FieldConstraints(fieldname, constraints)



# -*- coding: utf-8 -*-

#
# Unit tests for file functions from tdda.referencetest.checkfiles
#

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import division

import os
import unittest

from tdda.referencetest.checkfiles import FilesComparison
from tdda.referencetest.basecomparison import diffcmd


def refloc(filename):
    return os.path.join(os.path.dirname(__file__), 'testdata', filename)


def check(compare, values, filename, diff=False, actual_path=None):
    (code, errs) = compare.check_string_against_file(values, refloc(filename),
                                                     actual_path=actual_path,
                                                     create_temporaries=False)
    if not diff:
        errs = [e for e in errs if not e.startswith('    ' + diffcmd())]
        errs = [e for e in errs if not e.startswith('Compare with:')]
    return (code, errs)


class TestFiles(unittest.TestCase):
    def test_strings_against_files_ok(self):
        compare = FilesComparison()
        r1 = compare.check_string_against_file([], refloc('empty.txt'))
        r2 = compare.check_string_against_file('', refloc('empty.txt'))
        r3 = compare.check_string_against_file([''], refloc('empty.txt'))
        r4 = compare.check_string_against_file(['a single line'],
                                               refloc('single.txt'))
        self.assertEqual(r1, (0, []))
        self.assertEqual(r2, (0, []))
        self.assertEqual(r3, (0, []))
        self.assertEqual(r4, (0, []))

    def test_strings_against_files_fail(self):
        compare = FilesComparison()
        r1 = check(compare, ['x'], 'empty.txt')
        r2 = check(compare, 'x', 'empty.txt')
        r3 = check(compare, ['', ''], 'empty.txt')
        r4 = check(compare, ['the wrong text\n'], 'single.txt')
        r5 = check(compare, ['the wrong text\n'], 'single.txt', diff=True,
                   actual_path='wrong.txt')
        errs = ['Strings have different numbers of lines, '
                    'differences start at end of reference file',
                'Expected file %s' % refloc('empty.txt')]
        self.assertEqual(r1, (1, errs))
        self.assertEqual(r2, (1, errs))
        self.assertEqual(r3, (1, errs))
        self.assertEqual(r4, (1,
                              ['1 line is different, starting at line 1',
                               'Expected file %s' % refloc('single.txt')]))
        diff = '%s %s %s' % (diffcmd(), 'wrong.txt', refloc('single.txt'))
        self.assertEqual(r5, (1, ['1 line is different, starting at line 1',
                                  'Compare with:\n    %s\n' % diff]))

    def test_files_ok(self):
        compare = FilesComparison()
        r1 = compare.check_file(refloc('empty.txt'), refloc('empty.txt'))
        r2 = compare.check_file(refloc('single.txt'), refloc('single.txt'))
        r3 = compare.check_file(refloc('colours.txt'),
                                refloc('colours.txt'))
        self.assertEqual(r1, (0, []))
        self.assertEqual(r2, (0, []))
        self.assertEqual(r3, (0, []))

    def test_files_fail(self):
        compare = FilesComparison()
        r1 = compare.check_file(refloc('empty.txt'), refloc('single.txt'))
        r2 = compare.check_file(refloc('single.txt'), refloc('empty.txt'))
        r3 = compare.check_file(refloc('single.txt'),
                                refloc('colours.txt'))
        diff1 = '%s %s %s' % (diffcmd(),
                              refloc('empty.txt'), refloc('single.txt'))
        diff2 = '%s %s %s' % (diffcmd(),
                              refloc('single.txt'), refloc('empty.txt'))
        diff3 = '%s %s %s' % (diffcmd(),
                              refloc('single.txt'), refloc('colours.txt'))
        err1 = ('Files have different numbers of lines, '
               'differences start at end of actual file')
        err2 = ('Files have different numbers of lines, '
               'differences start at end of reference file')
        err3 = ('Files have different numbers of lines, '
               'differences start at line 1')
        self.assertEqual(r1, (1, [err1, 'Compare with:\n    %s\n' % diff1]))
        self.assertEqual(r2, (1, [err2, 'Compare with:\n    %s\n' % diff2]))
        self.assertEqual(r3, (1, [err3, 'Compare with:\n    %s\n' % diff3]))

    def test_file_removals(self):
        compare = FilesComparison()
        r = compare.check_file(refloc('removals.txt'),
                               refloc('ref.txt'),
                               remove_lines=['I am optional'])
        self.assertEqual(r, (0, []))

    def test_multiple_files_ok(self):
        compare = FilesComparison()
        r = compare.check_files([refloc('empty.txt'),
                                 refloc('single.txt'),
                                 refloc('colours.txt')],
                                [refloc('empty.txt'),
                                 refloc('single.txt'),
                                 refloc('colours.txt')])
        self.assertEqual(r, (0, []))

    def test_multiple_files_fail(self):
        compare = FilesComparison()
        r = compare.check_files([refloc('empty.txt'),
                                 refloc('single.txt'),
                                 refloc('colours.txt')],
                                [refloc('single.txt'),
                                 refloc('colours.txt'),
                                 refloc('colours.txt')])
        diff1 = '%s %s %s' % (diffcmd(),
                              refloc('empty.txt'), refloc('single.txt'))
        diff2 = '%s %s %s' % (diffcmd(),
                              refloc('single.txt'), refloc('colours.txt'))
        err1 = ('Files have different numbers of lines, '
               'differences start at end of actual file')
        err2 = ('Files have different numbers of lines, '
               'differences start at line 1')
        self.assertEqual(r, (2, [err1, 'Compare with:\n    %s\n' % diff1,
                                 err2, 'Compare with:\n    %s\n' % diff2]))

    def test_binary_files(self):
        compare = FilesComparison()
        r1 = compare.check_binary_file(refloc('single.txt'),
                                       refloc('single.txt'))
        r2 = compare.check_binary_file(refloc('single.txt'),
                                       refloc('double.txt'))
        r3 = compare.check_binary_file(refloc('double.txt'),
                                       refloc('single.txt'))
        r4 = compare.check_binary_file(refloc('single.txt'),
                                       refloc('single2.txt'))
        # single2.txt is deliberately not readable in text mode in python3
        self.assertEqual(r1, (0, []))
        if os.name != 'nt':
            # on Windows, the results will depend on hard-to-predict
            # factors such as how the sources were obtained from git,
            # whether it was a binary build or a source one, etc - so
            # for now we'll just check these on Unix.
            self.assertEqual(r2, (1, ['First difference at byte offset 14, '
                                      'actual length 14, expected length 28.']))
            self.assertEqual(r3, (1, ['First difference at byte offset 14, '
                                      'actual length 28, expected length 14.']))
            self.assertEqual(r4, (1, ['First difference at byte offset 2, '
                                      'both files have length 14.']))

    def test_removal_diffs(self):
        compare = FilesComparison()
        (code, msgs) = compare.check_file(refloc('removals.txt'),
                                          refloc('ref.txt'),
                                          remove_lines=['I am optional'])
        self.assertEqual(code, 0)
        self.assertEqual(msgs.lines, [])
        self.assertEqual(msgs.reconstructions[0].diff_actual,
                         ['This is a file containing some optional lines.',
                          "*** Here's one: I am optional"
                              "(|; but it's the only one; "
                              "the rest have been removed.)",
                          'And:',
                          "*** (Here's another one: "
                              "I am optional and I have some trailing stuff|)",
                          "And here's a line on its own:",
                          '*** (I am optional|)',
                          "That's all"])
        self.assertEqual(msgs.reconstructions[0].diff_expected,
                         ['This is a file containing some optional lines.',
                          "*** Here's one: I am optional"
                              "(|; but it's the only one; "
                              "the rest have been removed.)",
                          'And:',
                          "*** (Here's another one: "
                              "I am optional and I have some trailing stuff|)",
                          "And here's a line on its own:",
                          '*** (I am optional|)',
                          "That's all"])

    def test_removal_diffs_fail(self):
        compare = FilesComparison()
        (code, msgs) = compare.check_file(refloc('removals.txt'),
                                          refloc('ref.txt'),
                                          remove_lines=['line', 'And'])
        self.assertEqual(code, 1)
        self.assertEqual(len(msgs.lines), 7)
        self.assertEqual(msgs.lines[0],
                         'Files have different numbers of lines, '
                         'differences start at line 2')
        self.assertTrue(msgs.lines[1].startswith('Compare raw with:\n'))
        self.assertTrue(msgs.lines[2].startswith('Compare post-processed with:\n'))
        self.assertEqual(msgs.reconstructions[0].diff_actual,
                         ['*** This is a file containing some optional lines.',
                           # NEXT LINE IS A REAL DIFFERENCE
                          "Here's one: I am optional",
                          '*** And:',  # THIS IS REMOVED ON BOTH SIDES
                          "*** (|And here's a line on its own:)",
                           # NEXT LINE IS A REAL DIFFERENCE
                          "Here's another one: "
                              "I am optional and I have some trailing stuff",
                          "*** (And here's a line on its own:|)",
                           # NEXT TWO LINES ARE REAL DIFFERENCES
                          'I am optional',
                          "That's all"])
        self.assertEqual(msgs.reconstructions[0].diff_expected,
                         ['*** This is a file containing some optional lines.',
                           # NEXT LINE IS A REAL DIFFERENCE
                          "Here's one: I am optional; but it's the only one; "
                              "the rest have been removed.",
                          '*** And:',  # THIS IS REMOVED ON BOTH SIDES
                          "*** (|And here's a line on its own:)",
                           # NEXT LINE IS A REAL DIFFERENCE
                          "That's all",
                          "*** (And here's a line on its own:|)"])

    def test_ignore_substrings_diffs(self):
        compare = FilesComparison()
        (code, msgs) = compare.check_file(refloc('left.txt'),
                                          refloc('ref.txt'),
                                          ignore_substrings=['Here\'s one',
                                                             'And'])
        difflines = [
            'This is a file containing some optional lines.',
            "*** Here's one: "
                '('
                'And it will get ignored even if not optionally'
                '|'
                'I am optional; but it\'s the only one; '
                'the rest have been'
                ')'
                ' removed.',
            '*** And:'
                '('
                ' this line is different, unless you ignore '
                'the first word'
                '|'
                ')',
            "And here's a line on its own:",
            "That's all"
        ]
        self.assertEqual(code, 0)
        self.assertEqual(msgs.lines, [])
        self.assertEqual(msgs.reconstructions[0].diff_actual, difflines)
        self.assertEqual(msgs.reconstructions[0].diff_expected, difflines)

    def test_ignore_pattern_diffs(self):
        compare = FilesComparison()
        (code, msgs) = compare.check_file(refloc('left.txt'),
                                          refloc('ref.txt'),
                                          ignore_patterns=[
                                              '^.*opt...al.*$',
                                              '^.*[Aa][Nn][Dd].*$'
                                          ])
        difflines = [
            'This is a file containing some optional lines.',
            "*** Here's one: "
                '('
                'And it will get ignored even if not optionally'
                '|'
                "I am optional; but it's the only one; the rest have been"
                ')'
                ' removed.',
            '*** And:'
                '('
                ' this line is different, unless you ignore the first word'
                '|'
                ')',
            "And here's a line on its own:",
            "That's all"
        ]
        self.assertEqual(code, 0)
        self.assertEqual(msgs.lines, [])
        self.assertEqual(msgs.reconstructions[0].diff_actual, difflines)
        self.assertEqual(msgs.reconstructions[0].diff_expected, difflines)

    def test_ignore_pattern_diffs_fail(self):
        compare = FilesComparison()
        (code, msgs) = compare.check_file(refloc('left.txt'),
                                          refloc('ref.txt'),
                                          ignore_patterns=['^.*opt...al.*$'])
        difflines = [
            'This is a file containing some optional lines.',
            "*** Here's one: "
                '('
                'And it will get ignored even if not optionally'
                '|'
                "I am optional; but it's the only one; the rest have been"
                ')'
                ' removed.',
            '... placeholder ...',
            "And here's a line on its own:",
            "That's all"
        ]

        self.assertEqual(code, 1)
        self.assertEqual(len(msgs.lines), 6)
        self.assertEqual(msgs.lines[0],
                         '1 line is different, starting at line 3')
        self.assertEqual(msgs.lines[1][:8], 'Compare ')
        self.assertEqual(msgs.lines[2][:8], 'Compare ')
        self.assertEqual(msgs.lines[3], 'Note exclusions:')
        self.assertEqual(msgs.lines[4], '    ignore_patterns:')
        self.assertEqual(msgs.lines[5], '        ^.*opt...al.*$')

        difflines[2] = ('And: this line is different, '
                        'unless you ignore the first word')
        self.assertEqual(msgs.reconstructions[0].diff_actual, difflines)
        difflines[2] = 'And:'
        self.assertEqual(msgs.reconstructions[0].diff_expected, difflines)


if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals

import re
import regex
import unittest


NAME = 'FundaciÃ³ Rego'
PATTERN = '^[^\W0-9_]+ Rego$'

class TestRegex(unittest.TestCase):

    def testRE(self):
        self.assertIsNotNone(re.match(PATTERN, NAME, re.U))

    def testRegex(self):
        self.assertIsNotNone(regex.match(PATTERN, NAME, regex.U))


if __name__ == '__main__':
    unittest.main()





"""
Run all TDDA tests
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import sys

from tdda.referencetest import ReferenceTestCase

from tdda.constraints.testconstraints import *
from tdda.rexpy.testrexpy import *
from tdda.referencetest.tests.alltests import *


def testall(module=None, argv=None):
    ReferenceTestCase.main(module=module, argv=argv)


if __name__ == '__main__':
    testall(argv=sys.argv)



# -*- coding: utf-8 -*-
"""
test_using_referencepytest.py: A simple example of how to use
                               tdda.referencetest.referencepytest with pytest

Source repository: http://github.com/tdda/tdda

License: MIT

Copyright (c) Stochastic Solutions Limited 2019
"""

from generators import generate_string


def testExampleStringGeneration():
    actual = generate_string()
    with open('expected.html') as f:
        expected = f.read()
    assert actual == expected






# -*- coding: utf-8 -*-

"""
checkfiles.py: comparison mechanism for text files

Source repository: http://github.com/tdda/tdda

License: MIT

Copyright (c) Stochastic Solutions Limited 2016-2018
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
# from __future__ import unicode_literals

import os
import re
import sys
import tempfile
from collections import namedtuple

from tdda.referencetest.basecomparison import BaseComparison, Diffs, copycmd


BinaryInfo = namedtuple('BinaryInfo',
                        ('byteoffset', 'actualLen', 'expectedLen'))


class FilesComparison(BaseComparison):

    def check_strings(self, actual, expected,
                      actual_path=None, expected_path=None,
                      lstrip=False, rstrip=False,
                      ignore_substrings=None, ignore_patterns=None,
                      remove_lines=None,
                      preprocess=None, max_permutation_cases=0,
                      create_temporaries=True, msgs=None):
        """
        Compare two lists of strings (actual and expected), one-by-one.

            *actual*
                                is a list of strings.
            *expected*
                                is the expected list of strings.
            *actual_path*
                                is the path of the file where the actual
                                string originated (used for reporting errors).
            *expected_path*
                                is the path from which expected was read; it is
                                used to suggest a diff command to use on files
                                for strings failing the check.
            *lstrip*
                                if set to true, both strings are left-stripped
                                before the comparison is carried out.
                                Note: the stripping on a per-line basis.
            *rstrip*
                                if set to true, both strings are right-stripped
                                before the comparison is carried out.
                                Note: the stripping on a per-line basis.
            *ignore_substrings*
                                is an optional list of substrings; lines
                                containing any of these substrings will be
                                ignored in the comparison.
            *ignore_patterns*
                                is an optional list of regular expressions;
                                lines will be considered to be the same if
                                they only differ in substrings that match one
                                of these regular expressions. The expressions
                                should only include explicit anchors if they
                                need refer to the whole line.
                                Only the matched expression within the line is ignored;
                                any text to the left or right of the matched expression
                                must either be the same or be ignorable.
            *remove_lines*
                                is an optional list of substrings; lines
                                containing any of these substrings will be
                                completely removed before carrying out the
                                comparison. This is the means by which you
                                would exclude 'optional' content.
            *preprocess*
                                is an optional function that takes a list of
                                strings and preprocesses it in some way; this
                                function will be applied to both the actual
                                and expected.
            *max_permutation_cases*
                                is an optional number specifying the maximum
                                number of permutations allowed; if the actual
                                and expected lists differ only in that their
                                lines are permutations of each other, and
                                the number of such permutations does not
                                exceed this limit, then the two are considered
                                to be identical.
            *msgs*
                                is an optional Diffs object, where information
                                about differences will be appended; if not
                                specified, a new object will be created and
                                returned.

        Returns a tuple (failures, msgs), where failures is 1 if the lists
        differ and 0 if they are the same. The returned msgs is a list
        containing information about how they differed.
        """

        if msgs is None:
            msgs = Diffs()

        normalize = self.normalize_function(lstrip, rstrip)
        permutable = True
        failure_cases = []
        reconstruction = None
        format = None

        if preprocess:
            expected = preprocess(expected)
            actual = preprocess(actual)

        if actual and len(actual[-1]) == 0:
            actual = actual[:-1]
        if expected and len(expected[-1]) == 0:
            expected = expected[:-1]

        original_actual = actual
        original_expected = expected
        actual_map = {i: i for i in range(len(original_actual))}
        expected_map = {i: i for i in range(len(original_expected))}
        actual_removals = set()
        expected_removals = set()

        if remove_lines:
            actual_removals = set([i for i, a in enumerate(original_actual)
                                   if any(r in a for r in remove_lines)])
            expected_removals = set([i for i, a in enumerate(original_expected)
                                     if any(r in a for r in remove_lines)])
            actual = [a for i, a in enumerate(original_actual)
                      if i not in actual_removals]
            expected = [a for i, a in enumerate(original_expected)
                        if i not in expected_removals]

            # now build mappings from the after-removal line-numbers back to
            # the original line numbers, for both actual and expected
            iactual = 0
            for i in range(len(original_actual)):
                if i not in actual_removals:
                    actual_map[iactual] = i
                    iactual += 1
            iexpected = 0
            for i in range(len(original_expected)):
                if i not in expected_removals:
                    actual_map[iexpected] = i
                    iexpected += 1

        actual_ignored = set()
        expected_ignored = set()

        if len(actual) == len(expected):
            # Actual and expected now have same number of lines - so we just
            # need to check they are equivalent, line-by-line.
            diffs = [i for i in range(len(actual))
                     if normalize(actual[i]) != normalize(expected[i])]
            if diffs:
                wrong = self.wrong_content(diffs, actual, expected,
                                           actual_ignored, expected_ignored,
                                           actual_map, expected_map,
                                           failure_cases,
                                           max_permutation_cases,
                                           ignore_substrings=ignore_substrings,
                                           ignore_patterns=ignore_patterns)
                (first_error, ndiffs) = wrong
            else:
                first_error = None
                ndiffs = 0
        else:
            # Actual and expected have different numbers of lines, after
            # removal of optional lines.
            wrong = self.wrong_number(original_actual, original_expected,
                                      actual_ignored, expected_ignored,
                                      actual_removals, expected_removals,
                                      actual_map, expected_map,
                                      actual_path=actual_path,
                                      normalize=normalize,
                                      ignore_substrings=ignore_substrings,
                                      ignore_patterns=ignore_patterns)
            (first_error, ndiffs) = wrong
            permutable = False

        if (preprocess or actual_ignored or expected_ignored
                       or actual_removals or expected_removals
                       or (actual_path is None and ndiffs > 0)):
            norm_actual = [normalize(s) for s in original_actual]
            norm_expected = [normalize(s) for s in original_expected]
            reconstruction = self.reconstruct(norm_actual,
                                              norm_expected,
                                              actual_removals,
                                              expected_removals,
                                              actual_ignored,
                                              expected_ignored,
                                              format=format)
            msgs.add_reconstruction(reconstruction)

        if permutable and ndiffs > 0 and ndiffs <= max_permutation_cases:
            ndiffs = self.check_for_permutation_failures(failure_cases)

        if ndiffs > 0:
            if first_error:
                self.info(msgs, first_error)
            self.add_failures(msgs, reconstruction,
                              actual_path, expected_path,
                              ignore_substrings=ignore_substrings,
                              ignore_patterns=ignore_patterns,
                              remove_lines=remove_lines,
                              actual=actual, expected=expected,
                              preprocess=preprocess,
                              create_temporaries=create_temporaries)
        return (1 if ndiffs > 0 else 0, msgs)

    def wrong_content(self, diffs, actual, expected,
                      actual_ignored, expected_ignored,
                      actual_map, expected_map,
                      failure_cases, max_permutation_cases,
                      ignore_substrings=None, ignore_patterns=None):
        """
        Deal with the case where the actual and expected have the same
        number of lines (after removing 'removals'), so we need to examine
        those pairs of lines one-by-one to see what 'significant' differences
        there are.
        """
        ndiffs = len(diffs)
        first_error_line = None
        first_error = None
        if ndiffs > 0:
            compiled_patterns = self.compile_patterns(ignore_patterns)
            for i in diffs:
                # 'i' is a line-number in the after-removals datasets
                if self.can_ignore(actual[i], expected[i],
                                   ignore_substrings=ignore_substrings,
                                   compiled_patterns=compiled_patterns):
                    # a difference that can be ignored
                    ndiffs -= 1
                    actual_ignored.add(actual_map[i])
                    expected_ignored.add(expected_map[i])
                else:
                    # a difference that cannot be ignored
                    if first_error_line is None:
                        first_error_line = i + 1
                    if len(failure_cases) < max_permutation_cases:
                        failure_cases.append((i, actual[i], expected[i]))

            if first_error_line is not None:
                first_error = ('%d line%s different, starting at line %d'
                                % (ndiffs, 's are' if ndiffs != 1 else ' is',
                                   first_error_line))
        return (first_error, ndiffs)

    def wrong_number(self, original_actual, original_expected,
                     actual_ignored, expected_ignored,
                     actual_removals, expected_removals,
                     actual_map, expected_map, actual_path, normalize,
                     ignore_substrings=None, ignore_patterns=None):
        """
        Deal with the case where the actual and expected have different
        numbers of lines (after removing 'removal' lines).
        """
        compiled_patterns = self.compile_patterns(ignore_patterns)
        desc = 'file' if actual_path else 'string'
        iactual = 0
        iexpected = 0
        first_error_line = None
        ndiffs = 0
        nactual = len(original_actual)
        nexpected = len(original_expected)

        for i in range(min(nactual, nexpected)):
            removed = False
            if iactual in actual_removals:
                iactual += 1
                removed = True
            if iexpected in expected_removals:
                iexpected += 1
                removed = True
            if removed:
                continue
            actual_line = original_actual[iactual]
            expected_line = original_expected[iexpected]
            if normalize(actual_line) == normalize(expected_line):
                iactual += 1
                iexpected += 1
                continue
            if self.can_ignore(actual_line, expected_line,
                               ignore_substrings=ignore_substrings,
                               compiled_patterns=compiled_patterns):
                actual_ignored.add(actual_map[i])
                expected_ignored.add(expected_map[i])
                iactual += 1
                iexpected += 1
                continue
            first_error_line = 'line %d' % (iactual + 1)
            ndiffs += 1

        if first_error_line is None:
            if nactual > nexpected:
                first_error_line = 'end of reference file'
            else:
                first_error_line = 'end of actual %s' % desc

        first_error = ('%ss have different numbers of lines, '
                       'differences start at %s'
                       % (desc.title(), first_error_line))
        ndiffs = max(len(original_actual), len(original_expected))
        return (first_error, ndiffs)

    def compile_patterns(self, ignore_patterns):
        anchored_patterns = [('' if p.startswith('^') else '^(.*)')
                              + ('(%s)' % p)
                              + ('' if p.endswith('$') else '(.*)$')
                             for p in ignore_patterns or []]
        compiled_patterns = [re.compile(p) for p in anchored_patterns]
        return compiled_patterns

    def can_ignore(self, actual_line, expected_line,
                   ignore_substrings=None, compiled_patterns=None):
        """
        Can differences between this actual-line and this expected-line be
        ignored?
        """
        for substr in ignore_substrings or []:
            if substr in expected_line:
                # Note that this only looks in the 'expected' side.
                # This is deliberate, to cover the case where the expected
                # reference file contains concrete representations of things
                # that need to be ignored (like usernames, machine names etc),
                # where you can't express that more generally as any kind of
                # pattern.
                return True
        # not an ignorable substring line, so try patterns
        return self.check_patterns(compiled_patterns,
                                   actual_line, expected_line)

    def check_patterns(self, compiled_patterns, actual_line, expected_line):
        """
        Check a single pair of lines, taking regular-expressions into account.

        It expects the patterns to be in one of the following anchored forms:
            - a one-group pattern like ^(xxx)$
            - a two-group pattern like ^(.*)(xxx)$ or ^(xxx)(.*)$
            - a three-or-more-group pattern like ^(.*)(xxx)(.*)$ or ^(.*)(xxx)(yyy)(.*)$

        For a one-group pattern, it just needs to fully match both lines.

        For a two-group or three-group pattern, the (xxx) central part of
        the pattern needs to match for both lines, and the remaining (.*)
        parts both need to also match; they don't need to be identical, but
        they need to be 'equivalent' (i.e. by calling check_patterns() on
        these sub-parts, recursively).

        A fixed-string pattern (without any regular-expression components)
        needs to exactly match, on both actual and expected. So this is a
        STRONGER requirement than we have for ignore_substrings, since here
        we are requiring a match for BOTH the actual AND the expected.

        There might be an argument for having a mode where it's just the
        expected that contributes, but that's not currently provided.
        """
        if actual_line == expected_line:
            return True
        for pattern in compiled_patterns or []:
            mExpected = re.match(pattern, expected_line)
            if mExpected:
                mActual = re.match(pattern, actual_line)
                if not mActual:
                    continue
                if pattern.groups in (1, 2):
                    # matched a full-line expression
                    return True
                else:
                    actual_left = mActual.group(1)
                    expected_left = mExpected.group(1)
                    actual_right = mActual.group(pattern.groups)
                    expected_right = mExpected.group(pattern.groups)
                    if (self.check_patterns(compiled_patterns, actual_left, expected_left)
                            and self.check_patterns(compiled_patterns,
                                                    actual_right, expected_right)):
                        # the .* groups at start and end of both lines both match up, so
                        # this pair of lines can be ignored
                        return True
        return False

    def reconstruct(self, original_actual, original_expected,
                    actual_removals, expected_removals,
                    actual_ignored, expected_ignored, format):
        """
        Reconstruct variants of the actual and expected to generate
        two outputs that are different where we found differences, but
        are the same where we did not find differences (taking into
        account all of the various 'ignore' and 'remove' parameters)
        """
        rebuilt_actual = []
        rebuilt_expected = []
        iactual = iexpected = 0
        while (iactual < len(original_actual)
                   or iexpected < len(original_expected)):
            if iactual in actual_removals and iexpected in expected_removals:
                # lines which were removed from both sides
                marker = self.diff_marker(original_actual[iactual],
                                          original_expected[iexpected])
                rebuilt_actual.append(self.format_marker(marker, format))
                rebuilt_expected.append(self.format_marker(marker, format))
                iactual += 1
                iexpected += 1
            elif iactual in actual_removals:
                # line removed from just the left
                marker = self.diff_marker(original_actual[iactual], '')
                rebuilt_actual.append(self.format_marker(marker, format))
                rebuilt_expected.append(self.format_marker(marker, format))
                iactual += 1
            elif iexpected in expected_removals:
                # line removed from just the right
                marker = self.diff_marker('', original_expected[iexpected])
                rebuilt_actual.append(self.format_marker(marker, format))
                rebuilt_expected.append(self.format_marker(marker, format))
                iexpected += 1
            elif iactual >= len(original_actual):
                # fallen off the end of the left
                rebuilt_expected.append(original_expected[iexpected])
                iexpected += 1
            elif iexpected >= len(original_expected):
                # fallen off the end of the right
                rebuilt_actual.append(original_actual[iactual])
                iactual += 1
            elif (original_actual[iactual] == original_expected[iexpected]):
                # lines are the same, no differences
                rebuilt_actual.append(original_actual[iactual])
                rebuilt_expected.append(original_expected[iexpected])
                iactual += 1
                iexpected += 1
            elif iactual in actual_ignored or iexpected in expected_ignored:
                # lines are different, but differance has been ignored
                marker = self.diff_marker(original_actual[iactual],
                                          original_expected[iexpected])
                rebuilt_actual.append(self.format_marker(marker, format))
                rebuilt_expected.append(self.format_marker(marker, format))
                iactual += 1
                iexpected += 1
            else:
                # difference in line, not ignored, so it's a real difference!
                rebuilt_actual.append(original_actual[iactual])
                rebuilt_expected.append(original_expected[iexpected])
                iactual += 1
                iexpected += 1
        return Reconstruction(rebuilt_actual, rebuilt_expected)

    def diff_marker(self, left, right):
        """
        Generate a line to insert into the generated set of 'differences'
        which marks lines as:
                COMMON-PREFIX ( ONLY-IN-LEFT | ONLY-IN-RIGHT ) COMMON-SUFFIX
        """
        if left == right:
            return left    # nothing to do
        for i, (cl, cr) in enumerate(zip(list(left), list(right))):
            prefixend = i
            if cl != cr:
                break
        else:
            prefixend = min(len(left), len(right))
        for i, (cl, cr) in enumerate(zip(reversed(list(left[prefixend:])),
                                         reversed(list(right[prefixend:])))):
            postfixend = i
            if cl != cr:
                break
        else:
            postfixend = min(len(left[prefixend:]), len(right[prefixend:]))
        if postfixend > 0:
            middle = '(%s|%s)' % (left[prefixend:-postfixend],
                                  right[prefixend:-postfixend])
            marker = left[:prefixend] + middle + left[-postfixend:]
        else:
            middle = '(%s|%s)' % (left[prefixend:], right[prefixend:])
            marker = left[:prefixend] + middle
        return marker

    def format_marker(self, marker, format):
        """
        Format a 'diff marker' in a way that is appropriate for the kind
        of file it is being inserted into. Currently no special kinds of
        formats are treated specially, but it might be possible (though
        difficult) to do something more sensible for HTML files.
        """
        return '*** ' + marker

    def check_string_against_file(self, actual, expected_path,
                                  actual_path=None,
                                  lstrip=False, rstrip=False,
                                  ignore_substrings=None,
                                  ignore_patterns=None,
                                  remove_lines=None,
                                  preprocess=None, max_permutation_cases=0,
                                  create_temporaries=True, msgs=None):
        """
        Check a string (or list of strings) against the contents of a
        reference file.

        This is a wrapper around check_strings(), where the 'expected'
        strings are read from a file rather than being passed in explicitly.

        Other parameters are the same as for py:meth:`check_strings()`.

        The actual_path parameter is the pathname of the file that the
        actual string originally came from (if it came from a file at all;
        if not, it should be None).
        """
        if msgs is None:
            msgs = Diffs()
        try:
            with open(expected_path) as f:
                content = f.read()
                expected_ends_with_newline = content.endswith('\n')
                expected = content.splitlines()
        except IOError:
            self.info(msgs, 'Reference file %s not found.' % expected_path)
            self.add_failures(msgs, None, None, expected_path, actual=actual)
            return (1, msgs)

        if type(actual) in (list, tuple):
            actuals = actual
            actual_ends_with_newline = expected_ends_with_newline
        else:
            actuals = actual.splitlines()
            actual_ends_with_newline = actual.endswith('\n')
        mpc = max_permutation_cases
        (code, msgs) = self.check_strings(actuals, expected,
                                          actual_path=actual_path,
                                          expected_path=expected_path,
                                          lstrip=lstrip, rstrip=rstrip,
                                          ignore_substrings=ignore_substrings,
                                          ignore_patterns=ignore_patterns,
                                          remove_lines=remove_lines,
                                          preprocess=preprocess,
                                          max_permutation_cases=mpc,
                                          create_temporaries=create_temporaries,
                                          msgs=msgs)
        #if expected_ends_with_newline != actual_ends_with_newline:
        #    code = 1
        #    if actual_ends_with_newline:
        #        self.info(msgs, 'Actual string has unexpected newline at end')
        #    else:
        #        self.info(msgs, 'Actual string is missing newline at end')
        return (code, msgs)

    def check_file(self, actual_path, expected_path,
                   lstrip=False, rstrip=False,
                   ignore_substrings=None, ignore_patterns=None,
                   remove_lines=None,
                   preprocess=None, max_permutation_cases=0, msgs=None):
        """
        Check a pair of files, line by line, with optional
        ignore patterns (substrings) and optionally left-
        and/or right-stripping the contents of both files.

        This is a wrapper around check_strings(), where both the 'actual'
        and 'expected' strings are read from files rather than being passed
        in explicitly.

        Other parameters are the same as for :py:meth:`check_strings()`.

        """
        if msgs is None:
            msgs = Diffs()
        try:
            with open(expected_path) as f:
                content = f.read()
                expected_ends_with_newline = content.endswith('\n')
                expected = content.splitlines()
        except IOError:
            self.info(msgs, 'Reference file %s not found.' % expected_path)
            self.info(msgs,
                      'Initialize from actual content with:\n    %s %s %s'
                      % (copycmd(), actual_path, expected_path))
            return (1, msgs)
        try:
            with open(actual_path) as f:
                content = f.read()
                actual_ends_with_newline = content.endswith('\n')
                actuals = content.splitlines()
        except IOError:
            self.info(msgs, 'Actual file %s not found.'
                            % os.path.normpath(actual_path))
            self.add_failures(msgs, None, actual_path, expected_path)
            return (1, msgs)
        (code, msgs) = self.check_strings(actuals, expected,
                                          actual_path=actual_path,
                                          expected_path=expected_path,
                                          lstrip=lstrip, rstrip=rstrip,
                                          ignore_substrings=ignore_substrings,
                                          ignore_patterns=ignore_patterns,
                                          remove_lines=remove_lines,
                                          preprocess=preprocess,
                                          max_permutation_cases=
                                              max_permutation_cases,
                                          msgs=msgs)
        #if expected_ends_with_newline != actual_ends_with_newline:
        #    code = 1
        #    if actual_ends_with_newline:
        #        self.info(msgs, 'Actual string has unexpected newline at end')
        #    else:
        #        self.info(msgs, 'Actual string is missing newline at end')
        return (code, msgs)

    def check_files(self, actual_paths, expected_paths,
                    lstrip=False, rstrip=False,
                    ignore_substrings=None, ignore_patterns=None,
                    remove_lines=None,
                    preprocess=None, max_permutation_cases=0, msgs=None):
        """
        Compare a list of files against a list of reference files.

        It compares all the files pair-wise from the two lists, and then
        reports on any differences. This is different from calling
        check_file() separately for each pair, which will stop as soon
        as the first difference is found.

        Other parameters are the same as for :py:meth:`check_strings()`.

        """
        failures = 0
        if msgs is None:
            msgs = Diffs()
        for (actual_path, expected_path) in zip(actual_paths, expected_paths):
            try:
                r = self.check_file(actual_path, expected_path,
                                    ignore_substrings=ignore_substrings,
                                    ignore_patterns=ignore_patterns,
                                    remove_lines=remove_lines,
                                    preprocess=preprocess,
                                    lstrip=lstrip, rstrip=rstrip,
                                    max_permutation_cases=max_permutation_cases,
                                    msgs=msgs)
                (n, msgs) = r
                failures += n
            except Exception as e:
                self.info(msgs, 'Error comparing %s and %s (%s %s)'
                                % (os.path.normpath(actual_path),
                                   expected_path,
                                   e.__class__.__name__, str(e)))
                failures += 1
        return (failures, msgs)

    def check_binary_file(self, actual_path, expected_path, msgs=None):
        """
        Check a pair of binary files.
        """
        if msgs is None:
            msgs = Diffs()
        try:
            with open(expected_path, 'rb') as f:
                expected = f.read()
        except IOError:
            self.info(msgs, 'Reference file %s not found.' % expected_path)
            self.info(msgs,
                      'Initialize from actual content with:\n    %s %s %s'
                      % (copycmd(), actual_path, expected_path))
            return (1, msgs)
        try:
            with open(actual_path, 'rb') as f:
                actual = f.read()
        except IOError:
            self.info(msgs, 'Actual file %s not found.'
                            % os.path.normpath(actual_path))
            self.add_failures(msgs, None, actual_path, expected_path)
            return (1, msgs)

        if expected == actual:
            return (0, msgs)
        minlen = min(len(expected), len(actual))
        if expected[:minlen] == actual[:minlen]:
            boff = minlen
        else:
            boff = 0
            while boff < minlen:
                if expected[boff] != actual[boff]:
                    break
                boff += 1
        self.add_failures(msgs, None, actual_path, expected_path,
                          actual=actual, expected=expected,
                          binaryinfo=BinaryInfo(boff,
                                                actualLen=len(actual),
                                                expectedLen=len(expected)),
                          create_temporaries=False)
        return (1, msgs)

    def check_for_permutation_failures(self, failure_cases):
        """
        Check a collection of actual/reference comparison lines, to
        see if they are different only because one is a permutation of
        the other.
        """
        actuals = [f[1] for f in failure_cases]
        expecteds = [f[2] for f in failure_cases]
        if self.verbose and self.print_fn:
            self.print_fn('W(%s)' %len(actuals), file=sys.stderr)
        if sorted(actuals) == sorted(expecteds):
            return 0
        else:
            return len(actuals)

    def normalize_function(self, left, right):
        """
        Return the appropriate function for stripping a string,
        with left and right being booleans that specify whether
        to strip on the two sides.
        """
        if left and right:
            return lambda s: s.strip()
        elif left:
            return lambda s: s.lstrip()
        elif right:
            return lambda s: s.rstrip()
        else:
            return lambda s: s

    def add_failures(self, msgs, reconstruction, actual_path, expected_path,
                     ignore_substrings=None, ignore_patterns=None,
                     remove_lines=None,
                     preprocess=None, actual=None, expected=None,
                     binaryinfo=None, create_temporaries=True):
        """
        Build a list of messages describing the way in which two files are
        different, and construct an appropriate 'diff' command.
        """
        binary = binaryinfo is not None
        assert not (binary and reconstruction)

        commonname = None
        differ = None
        raw_actual_path = actual_path
        raw_expected_path = expected_path

        if create_temporaries:
            if raw_actual_path and raw_expected_path:
                commonname = os.path.split(actual_path)[1]
            else:
                if raw_actual_path:
                    commonname = os.path.split(actual_path)[1]
                elif raw_expected_path:
                    commonname = os.path.split(expected_path)[1]
                else:
                    commonname = 'file'
                if expected is not None and not raw_expected_path:
                    # no raw expected file, so write it
                    tmpExpectedPath = os.path.join(self.tmp_dir,
                                                   'expected-raw-' + commonname)
                    raw_expected_path = tmpExpectedPath
                    self.write_file(raw_expected_path, expected)
                if actual is not None and not raw_actual_path:
                    # no raw actual file, so write it
                    tmpActualPath = os.path.join(self.tmp_dir,
                                                 'actual-raw-' + commonname)
                    raw_actual_path = tmpActualPath
                    self.write_file(raw_actual_path, actual)

        if raw_actual_path and raw_expected_path:
            raw = 'raw' if (preprocess or reconstruction) else None
            differ = self.compare_with(raw_actual_path, raw_expected_path,
                                       qualifier=raw, binary=binary)

        if not actual_path or not expected_path:
            if expected_path:
                self.info(msgs, 'Expected file %s' % expected_path)
            elif actual_path:
                self.info(msgs,
                          'Actual file %s' % os.path.normpath(actual_path))
            elif not create_temporaries:
                self.info(msgs, 'No files available for comparison')

        if differ:
            self.info(msgs, differ)

        if reconstruction and create_temporaries:
            # show diffs after ignores and removals have been collapsed
            if differ:
                differ = '***\n' + differ + '***\n\n'
            diffActual = os.path.join(self.tmp_dir, 'actual-' + commonname)
            diffExpected = os.path.join(self.tmp_dir, 'expected-' + commonname)
            guide = expected_path or actual_path
            self.write_file(diffActual,
                            (differ or '') + reconstruction.actual_lines(),
                            guide=guide)
            self.write_file(diffExpected,
                             (differ or '') + reconstruction.expected_lines(),
                            guide=guide)
            self.info(msgs, self.compare_with(diffActual, diffExpected,
                                              qualifier='post-processed'))

        if ignore_substrings or ignore_patterns or remove_lines:
            self.info(msgs, 'Note exclusions:')
        if ignore_substrings:
            self.info(msgs, '    ignore_substrings:')
            for substr in ignore_substrings:
                self.info(msgs, '        ' + substr)
        if ignore_patterns:
            self.info(msgs, '    ignore_patterns:')
            for pattern in ignore_patterns:
                self.info(msgs, '        ' + pattern)
        if remove_lines:
            self.info(msgs, '    remove_lines:')
            for substr in remove_lines:
                self.info(msgs, '        ' + substr)
        if binary:
            if binaryinfo.actualLen == binaryinfo.expectedLen:
                lengthinfo = 'both files have length %d' % binaryinfo.actualLen
            else:
                lengthinfo = ('actual length %d, expected length %d'
                              % (binaryinfo.actualLen, binaryinfo.expectedLen))
            self.info(msgs, 'First difference at byte offset %d, %s.'
                      % (binaryinfo.byteoffset, lengthinfo))

    def write_file(self, filename, contents, guide=None):
        """
        Write contents out to a file, optionally taking guidance from an
        existing file as to whether to a newline at the end or not.
        """
        if type(contents) in (list, tuple):
            contents = '\n'.join(contents)
        with open(filename, 'w') as f:
            f.write(contents)
            if guide:
                with open(guide) as fg:
                    lastline = None
                    for line in fg.read():
                        lastline = line
                    if lastline and lastline.endswith('\n'):
                        f.write('\n')


class Reconstruction(object):
    """
    Class for representing 'reconstructions' of the differences between
    a pair of files, in the form of lists of lines from each, where
    ignored and removed items have been 'collapsed' so that the remaining
    differences are just the ones that the comparison considers to be
    actually 'different'.
    """
    def __init__(self, diff_actual, diff_expected):
        self.diff_actual = diff_actual
        self.diff_expected = diff_expected

    def actual_lines(self):
        return '\n'.join(self.diff_actual)

    def expected_lines(self):
        return '\n'.join(self.diff_expected)





# -*- coding: utf-8 -*-
"""
exercise1/test_all.py: A simple exercise to show tagging tests

Source repository: http://github.com/tdda/tdda

License: MIT

Copyright (c) Stochastic Solutions Limited 2019
"""

import time

from tdda.referencetest import ReferenceTestCase

from generators import generate_string


class TestQuickThings(ReferenceTestCase):

    def testExampleStringGeneration(self):
        actual = generate_string()
        self.assertStringCorrect(actual, 'expected.html')

    def testZero(self):
        self.assertIsNone(None)


class TestSuperSlowThings(ReferenceTestCase):

    def testOne(self):
        time.sleep(1)
        self.assertEqual(1, 1)

    def testTwo(self):
        time.sleep(2)
        self.assertEqual(2, 2)

    def testThree(self):
        time.sleep(3)
        self.assertEqual(3, 3)


if __name__ == '__main__':
    ReferenceTestCase.main()




# elements_detect_118_against_92.py

from __future__ import print_function
import pandas as pd

from tdda.constraints.pd.constraints import detect_df

df = pd.read_csv('testdata/elements118.csv')
print(detect_df(df, 'elements92.tdda', outpath='elements118_detect.csv',
                per_constraint=True, output_fields=[]))




# accounts_detect_25k_against_1k.py

from __future__ import print_function
import pandas as pd

from tdda.constraints.pd.constraints import detect_df

df = pd.read_csv('testdata/accounts25k.csv')
print(detect_df(df, 'accounts1k.tdda', outpath='accounts25k_detect.csv',
                per_constraint=True, output_fields=[]))




# -*- coding: utf-8 -*-
"""
conftest.py: example pytest configuration for tdda.referencetest

Source repository: http://github.com/tdda/tdda

License: MIT

Copyright (c) Stochastic Solutions Limited 2016-2018
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tdda.referencetest.pytestconfig import (pytest_addoption,
                                             pytest_collection_modifyitems,
                                             set_default_data_location,
                                             ref)


set_default_data_location('../reference')




# -*- coding: utf-8 -*-

#
# Unit tests for base ReferenceTest class functionality
#

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import division

import os
import unittest

from tdda.referencetest.referencetest import ReferenceTest


class TestReferenceTest(unittest.TestCase):
    def testDefaultLocations(self):
        class ClassA(ReferenceTest):
            pass
        class ClassB(ClassA):
            pass
        class ClassC1(ClassB):
            pass
        class ClassC2(ClassB):
            pass
        class classZ(ReferenceTest):
            pass

        a = ClassA(None)
        b = ClassB(None)
        c1 = ClassC1(None)
        c2 = ClassC2(None)

        self.assertEqual(a._resolve_reference_path('x'), 'x')
        self.assertEqual(b._resolve_reference_path('x'), 'x')
        self.assertEqual(c1._resolve_reference_path('x'), 'x')
        self.assertEqual(c2._resolve_reference_path('x'), 'x')

        ClassA.set_default_data_location('t1')
        ClassB.set_default_data_location('t2')
        ClassC1.set_default_data_location('t3')

        a = ClassA(None)
        b = ClassB(None)
        c1 = ClassC1(None)
        c2 = ClassC2(None)

        self.assertEqual(a._resolve_reference_path('x'),
                         os.path.join('t1', 'x'))
        self.assertEqual(b._resolve_reference_path('x'),
                         os.path.join('t2', 'x'))
        self.assertEqual(c1._resolve_reference_path('x'),
                         os.path.join('t3', 'x'))
        self.assertEqual(c2._resolve_reference_path('x'),
                         os.path.join('t2', 'x'))


if __name__ == '__main__':
    unittest.main()




try:
    import pandas as pd
except ImportError:
    pd = None

if pd is not None:
    from tdda.constraints.pd.constraints import (discover_df,
                                                 verify_df,
                                                 detect_df)

from tdda.constraints.db.constraints import (discover_db_table,
                                             verify_db_table,
                                             detect_db_table)




# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
"""1.0.02
Improved documentation to use shorter import forms.

Also bumped version number.

Also planning to tag this with a version to see whether that encourages
readthedocs to add the version properly.

29.05.2018 1.0.03
Add pushv (for maintainer use only) and this CHANGES.py file.

29.05.2018 1.0.04
Correctly cast strings to native strings in check_shell_output.

29.05.2018 1.0.05
Tests for feather are now ignored if pmmif/feather are not installed.

29.05.2018 1.0.06
Windows fixes.

31.05.2018 1.0.07
Fix for UTF-8 encoding in subprocesses on Windows.

Also added documentation on setting up fonts on Windows to display
ticks and crosses correctly.

6.06.2018 1.0.08
Improved metadata in TDDA files.

11.06.2018 1.0.09
Add support for postloadhook and postdicthook.

25.07.2018 1.0.10
Minor bug fixes.

 - Fuzzy verification of 'min' constraint was printing out the type of the col.
 - Typo in documentation of the properties available in a Verification result.
 - Verification of 'rex' constraint was not checking that the field exists.

26.07.2018 1.0.11
Fixed bug in pandas detect; it wasn't detecting min/max length constraints.

Also fixed issue with pandas CSV file reader; it has problems reading files
that have stuttered quotes and which also have escaped content. It now
notices if that has gone wrong, and has another try.

26.07.2018 1.0.12
minor refactoring and comments

1.08.2018 1.0.13
Now accepts 'false' as a valid specification value for no-duplicates.

Also improved the error message you get if you provide invalid specifications.
Also suppressed pandas warning about nanoseconds on conversion.

1.08.2018 1.0.14
Ignore epsilon for date min/max (rather than crashing).

This probably isn't ideal, but to make it better we'd need to decide
on what epsilon should mean for dates, which is not obvious.

13.08.2018 1.0.15
Change use of re.U (UNICODE) to re.UNICODE | re.DOTALL.

This means that dot actually matches any character.
The Python documentation claims that without this any character except
newline is matched, but it appears that some other characters are
also not matched without re.DOTALL, including a non-breaking space (0x80).

18.09.2018 1.0.16
referencetest class now exports TaggedTestLoader for convenience.

18.09.2018 1.0.17
TestLoader now takes an optional 'printer' parameter to control how -0 works.

18.09.2018 1.0.18
Renamed tdda.referencetest method assertFileCorrect -> assertTextFileCorrect.

The original name is still available for backwards compatibility, but
is deprecated.

19.09.2018 1.0.19
Added tdda.referencetest method assertBinaryFileCorrect.
Also fixed some issues with tdda.referencetest ignore_patterns method.

19.09.2018 1.0.20
Added test files so that all the new tests will pass.

28.02.2019 1.0.21
Changed 'test' command to from 'this' Python rather than a new Python.

Previously, the 'test' command (run with 'tdda test') ran in a subshell
using an os.system call. In order to do this, we used sys.executable()
to find (ostensibly) the path to the running version of Python.

It transpires, however, that sys.executable() does not always return
the path to the running Python...and indeed sometimes returns a path
that does not exist. We now, therefore, simply run the tests in the
current Python, by calling the (new) function 'testall', now used
by main() in tdda.testtdda.py.

13.03.2019 1.0.22
Fix for fuzzy min/max comparison on date columns, previously crashing.

Also fixed issue with problems, on some platforms, with the 'tdda examples'
command, for non-ascii example files. It had already had support for doing
this, but it wasn't always working. Now it's much more bullet-proof.

Also, it no longer generates warnings for unrecognised keys in the input
.tdda file (on verify and detect), for keys with names that begin with '#'.
This gives a mechanism for 'commented out' items in the file (which you
could always do, but you'd get annoying warnings, which now you no longer
get).

Also added some examples of 'detect' in the constraints examples and README.

Also added some examples for the 'accounts' banking datasets for constraints,
and included this new dataset as part of the standard examples you get if
you run 'tdda examples'.

Also fixed typo in link to 1-pager pdf in the documentation.

Added tests for accounts-based constraints examples.
Also, copying examples now unzips the file.

18.03.2019 1.0.23
Fixed tests to pass under Python 3.7.2.

The standard re.escape method changed after python version 3.7.1 so that it
only escapes characters that would otherwise be treated as having special
meaning within a regular expression. So the tests now need to allow BOTH
outputs to be considered correct. It's less easy than you'd imagine to work
out exactly what subversion the behaviour changed in.

Small fix to one test to ensure it still works in Pythons 2.7 & 3.6.

Fixed reftest example to use ignore_substrings rather than ignore_patterns
(as a result of the change to semantics of ignore_patterns in 1.0.19).

Added API examples for reference tests using accounts data.

------------------------- dev branch -------------------------
Refactored reporting of differences for files and strings, to take
better account of 'ignore' and 'remove' parameters. It now builds a 'diff'
that will only appear 'different' for lines that are REALLY different, after
any removals and ignores have been collapsed. That diff can also include an
embedded 'raw' diff, which will show ALL the differences, but the main focus
is on showing 'what is different that should not be'.

Also improved reporting of differences when there are different numbers
of lines. It now says what line at which the (effective) differences start
(taking into account removals and ignores).

Also improved reporting of differences when there is no 'actual' file to
compare (just a string in memory, with assertStringCorrect). It now produces
a temporary file containing the 'actual' contents, if there are errors to
report, so that a 'diff' command can be generated in the same way as it does
when there is a file already available.

The ignore_patterns parameter is now treated slightly more strictly than
before, and has had its documentation improved. If you provide an unanchored
regular expression pattern, it now requires that both sides match that
pattern, but it ALSO now requires that the remaining parts, to the left and
right of where the pattern matched, be 'equivalent' if the line is to be
ignored. The two 'left parts' must be equivalent, and the two 'right parts'
must also be equivalent. This 'equivalence' is checked by (recursively)
applying the same logic to these sub-parts.

The ignore_substrings parameter is now treated slightly more strictly than
before. Previously, a line was ignored if it contained any of the ignorable
substrings, in either the actual or the expected. That meant that lines in
the actual data might be being unexpectedly ignored if they start to include
such strings, which probably was not the intent of the test at all. Now,
ignorable substrings only refer to the *expected* data (which is fixed for
the test, and you know exactly what is in it and what is not).

Add comment=None to all constraint constructors.

This allows comments to be added to individual constraints by using
a dictionary for constraint values with keys "value" (for the constraint
value) and "comment" as an ignored string.

Added better support for unittest single-letter command-line option flags,
allowing things like -W, -1 and -0 to be grouped (with themselves, and with
standard unittest options). This means that passing grouped flags such as -1v
or -1W now works as expected with referencetest tests under unittest.

Relaxed limitation that ignore_patterns regex expressions in referencetest tests
couldn't include grouped expressions. Now they can (which is very useful for when
you want to use alternations in expressions, like (a|b)).

Better reporting of exclusions when checkfiles reports test failures.

Updated reference test tests.

Reporting differences when there are different numbers of lines now improved.

Fixed problem with set_defaults, where print_fn setting wasn't working.

Add -q/--quote flag to rexpy CLI to quote output strings.

Fixed problems with exceptions being raised for constraints that couldn't
apply because of type mis-match. (e.g. if a field is supposed to be a string,
and has min/max length constraints, or a rex constraint, then these should
fail when verified in an environment when it turns out not to be a string
any more; but it was raising an exception instead). Similar for things like
sign constraints, where it's expecting a numeric field, but it should cope
(and fail the constraint rather than blowing up) if the field is not numeric.

Added some 'convenience' imports at the root of the tdda module. So now
you can just import 'tdda', and then refer to things within tdda.constraints
or tdda.referencetest or tdda.rexpy. Previously you had to individually
import each submodule.

Updated MongoDB constraint support, which had rusted.

Updated constraint support for Pandas dataframes to include datetime.date
columns. Previously it only supported datetime.datetime (and raised an
exception if given a date-only).

Extended Pandas verify_df and detect_df so that they can accept an in-memory
dictionary to represent the constraints to verify/detect, as well as accepting
a path for a .tdda file. This means that, when using the tdda.constraints API,
you can now use the library to verify/detect against constraints that are
stored in other places, not just in the filesystem. (E.g. S3, version control,
etc).

------------------------- end of dev branch -------------------------

2.04.2019 1.0.24
Merged dev branch into master.

All changes listed above are now included in master.

10.04.2019 1.0.25
Fix in rexpy code for computing incremental coverage, to deal with some
edge cases that were hanging. It wasn't taking account of the possibility
that a *subset* of the expressions found could give 100% coverage, and
therefore some of the other expressions aren't contributing anything.

Fix for output regeneration in referencetests having rusted.
It was missing the new method _normalize_whitespace.

Merged dev branch into master again.

Fixed various documentation and release notes issues, including expanding
and updating the list of drivers supported and required for various
databases and other data sources.

Minor changes to tests to make them report more accurately and completely
what's going on when there are skipped tests.

Added boilerplate tests for MongoDB, but these are currently disabled
since they cannot yet work entirely correctly.

30.05.2019 1.0.26
Better JSON generation and protect against unfathomable user.

Tweaked the JSON produced for TDDA files:
  1. ensure no trailing whitespace, even in Python2
  2. don't force ASCII, so that accented characters, greek characters etc.,
     come out in UTF-8, rather than as character codes.

Also protect the lookup of the current user in a try ... except
and set to the empty string if it fails. This addresses issue #18,
whereby (poorly configured?) Docker containers with no non-root users
can cause getpass.getuser() to fail with an error. Whether or not
the software should need to deal with this, there seems little harm
in setting an empty user when the alternative is crashing, so that's
what now happens.

3.06.2019 1.0.27
Less excessive escaping (Python3-like even in Python2)

4.06.2019 1.0.28
Fixed (nasty) bug in escaping (above). Updated/improved tests.

Now that we do regular expression escapinge in (almost) the same way in
Python2 and Python3, we can remove a lot of nasty tests that have
two versions of the results.

5.06.2019 1.0.29
Updated bracket generatation ("character classes: [...]").

We now use almost no escaping for characters in character classes,
instead using special rules to force
  - close square bracket ("]") to the start
  - dash (hyphen "-") to the end
  - carat ("^") away from the start
Only backslash is now normally escaped in character classes.

There is also embryonic provision for new dialects "javascript" and "ruby"
to force escaping of "]" in character classes, since they don't appear to
obey the normal rules, but the embryonic provision isn't really used by
anything (or available outside the API) at this point.

Many tests have been updated as a result. Many tests also now check,
to a greater extent than was previously the case, that the regular expressions
generated actually actually match the strings used to generate them and
(equally relevantly) don't match certain strings you might worry they would
match if you weren't fully "au fait" with the details of regex rules!

13.06.2019 1.0.30
Fixed problem with the 'tdda test' command. It was failing with:
    AttributeError: module 'tdda.testtdda' has no attribute 'test'

Also simplified the list of characters to NOT be specially escaped for rexpy,
so it's now just the same across all python versions.

25.10.2019 1.0.31
Constraint verify/detect of regular expressions now handles '.' better.

It wasn't set to allow newlines to match '.', meaning that regular expressions
matching text with newlines in it wasn't always matching as you might expect.

Reftest Exercises added.
"""



# accounts_verify_25k_against_1k_feather.py

from __future__ import print_function
import pandas as pd

from tdda.constraints.pd.constraints import verify_df

df = pd.read_csv('testdata/accounts25k.csv')
verification = verify_df(df, 'testdata/accounts1k.tdda')

print('Basic Verification:')
print(verification)
print('\n')
print('Verification DataFrame:')
dfv = verification.to_frame()
print(dfv)




# -*- coding: utf-8 -*-
"""
exercise1/test_all.py: A simple example of how to use
                       tdda.referencetest.ReferenceTestCase.

Source repository: http://github.com/tdda/tdda

License: MIT

Copyright (c) Stochastic Solutions Limited 2019
"""

import unittest

from generators import generate_string


class TestFileGeneration(unittest.TestCase):
    def testExampleStringGeneration(self):
        actual = generate_string()
        with open('expected.html') as f:
            expected = f.read()
        self.assertEqual(actual, expected)


if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals

import re
import regex
import unittest


class TestRegex(unittest.TestCase):
    s = 'FundaciÃ³ Rego'
    R = '^[^\W0-9_]+ Rego$'

    self.assertIsNotNone(re.match(R, s, re.U))

    self.assertIsNotNone(regex.match(R, s, regex.U))


if __name__ == '__main__':
    unittest.main()


